"""
File Purpose: tools related to multiprocessing
"""

import multiprocessing as mp
import pickle
import time  # for mptest functions like sleep

import numpy as np

from .arrays import nest_shape
from .docs_tools import format_docstring
from .iterables import (
    is_iterable, is_dictlike,
    Container, ContainerOfList, ContainerOfArray, ContainerOfDict,
)
from .properties import alias, simple_property
from .sentinels import RESULT_MISSING, UNSET, Sentinel
from .timing import TimeLimit, ProgressUpdater
from ..errors import InputError, InputConflictError

from ..defaults import DEFAULTS


_paramdocs_tasks = {
    'kw': '''dict
        kwargs for task will be task.kw, but updated with kw.
        E.g. if task.kw = {'x': 1}, and kw = {'y': 2}, --> task called with x=1, y=2.''',
    'reset': '''bool
        whether to reset self.result to all RESULT_MISSING, before starting this operation.''',
    'skip_done': '''bool
        whether to skip tasks that already have a result (i.e. self.result[idx] != RESULT_MISSING).''',
    'ncpu': '''None or int
        max number of cpus to use for multiprocessing.
        None --> use multiprocessing.cpu_count()
        int --> use this value. if 0 or 1, do not use multiprocessing here.
        Note: will actually use min(ncpu, number of calls to be made);
            e.g. if ncpu=4 but len(arg_kw_tuples)=2, will only use 2 cpus.''',
    'timeout': '''None or int
        max duration, in seconds. Must be None or integer (due to limitations of signal.alarm method)
        None --> no time limit.
        Note: if time_limit is reached, will raise a TimeoutError and save the result so far.
            (in this case, any not-yet-calculated values will each be RESULT_MISSING.)''',
    'f': '''callable
        The function to be called during task evaluation.
        if using ncpu > 1, f must be compatible with multiprocessing, i.e. it must be pickleable,
            e.g. it should be defined at the top level of a module or as a class method.''',
    'ncoarse': '''int
        if >1, group tasks into groups of size ncoarse before performing them.''',
    'print_freq': '''None, or number (possibly negative or 0)
        >0 --> Minimum number of seconds between progress updates.
        =0 --> print every progress update.
        <0 --> never print progress updates.
        None --> use DEFAULTS.PROGRESS_UPDATES_PRINT_FREQ''',
    'assign_task_idx': '''bool
        whether to assign task.i for each task, based on its position in self.''',
    'printable_process_name': '''None or str
        include in progress update printouts. If None, use a reasonable default.''',
    'errors_ok': '''bool, Exception type, or tuple of Exception types
        whether it is okay for some tasks to produce certain errors.
        False --> crash if any task crashes. Equivalent to errors_ok=().
        True --> except Exception (not BaseException, though). Equivalent to errors_ok=Exception.
        Exception type or tuple --> except this type (or these types, if tuple).''',
    'result_missing': '''any object
        result to record for tasks which crash (if errors_ok!=False). Default RESULT_MISSING.''',
}

''' ----------------------------- Helper functions ----------------------------- '''

def is_task_element(x):
    '''returns whether x is a Task-like element.
    i.e. x is a callable, or len(x)=0, or x iterable but x[0] not iterable nor callable.
    '''
    if callable(x):
        return True
    elif is_iterable(x):
        return len(x)==0 or (not callable(x[0]) and not is_iterable(x[0]))
    else:
        raise TypeError(f'is_task_element(x) expected x to be callable or iterable, but got x={x}')

def check_pickle(x):
    '''checks that x, is pickleable, by pickling then unpickling.
    Returns result of unpickling. Useful for debugging.
    '''
    dump = pickle.dumps(x)
    load = pickle.loads(dump)
    return load

def copy_via_pickle(x):
    '''returns copy of x via pickle.loads(pickle.dumps(x)).'''
    dump = pickle.dumps(x)
    return pickle.loads(dump)


''' ----------------------------- Task ----------------------------- '''

@format_docstring(**_paramdocs_tasks)
class Task():
    '''a single task, i.e. a function, args, & kwargs.
    Calling self will perform the task, i.e. self.result=f(*args, **kw), and return result.

    Alternatively, calling self.apply_async(pool) enables to perform task in parallel,
        using the muliprocessing module.

    f: {f}
    args: iterable of args or non-iterable object.
        if iterable, will be interpreted as the list of args.
        if non-iterable, will use args = [args].
    i: any object
        optionally, index for this task.
        Useful when task appears inside a TaskContainer.
    '''
    # [TODO] is there any non-negligible time cost to making Task objects instead of passing tuples?
    def __init__(self, f, args=None, kw=None, *, i=None):
        self.f = f
        self.args = () if args is None else (args if is_iterable(args) else [args])
        self.kw = dict() if kw is None else kw
        self.i = i
        if not is_dictlike(self.kw):
            errmsg = f'{type(self).__name__}(...) expected kw to be dict-like, but got kw={kw}'
            raise TypeError(errmsg)

    def new_with(self, f=UNSET, args=UNSET, kw=UNSET, i=UNSET):
        '''return a new Task, with the given inputs. Use value from self for UNSET inputs.
        E.g. Task(f, [5,6]).new_with(f=g, kw=dict(y=7)) --> Task(g, [5,6], dict(y=7))
        '''
        cls = type(self)
        f = self.f if f is UNSET else f
        args = self.args if args is UNSET else args
        kw = self.kw if kw is UNSET else kw
        i = self.i if i is UNSET else i
        return cls(f, args, kw, i=i)

    @classmethod
    def implied(cls, task_info, *, force_kw=True, **kw_init):
        '''return the Task implied from task_info.
        task_info: Task, iterable, or callable.
            Task --> return task_info, unchanged (unless force_kw and len(kw_init)>0).
                     if task_info is not an instance of cls but it is a Task, raise TypeError.
            iterable --> return Task(*task_info).
            callable --> return Task(task_info).
        force_kw: bool
            applies if kw_init provided and task_info is a Task...
                True --> return task_info.new_with(**kw_init)
                False --> return task_info, unchanged.
        kw_init: pass to Task(..., **kw_init) if task_info is not already a Task.
        '''
        if isinstance(task_info, Task):
            if isinstance(task_info, cls):
                if force_kw and len(kw_init)>0:
                    return task_info.new_with(**kw_init)
                else:
                    return task_info
            else:
                raise TypeError(f'{cls.__name__}.implied(t) expected t subtype of {cls.__name__}, but got t={task_info}.')
        elif is_iterable(task_info):
            return cls(*task_info, **kw_init)
        elif callable(task_info):
            return cls(task_info, **kw_init)
        else:
            errmsg = f'{cls.__name__}.implied(t) expected {cls.__name__}, iterable, or callable, but got t={task_info}.'
            raise TypeError(errmsg)

    # # # PERFORM THE TASK # # #
    @format_docstring(**_paramdocs_tasks, sub_ntab=1)
    def __call__(self, **kw):
        '''perform the task, i.e. return f(*args, **kw).

        kw: {kw}
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        self.result = self.f(*self.args, **self.get_kw(kw=kw))
        return self.result

    @format_docstring(**_paramdocs_tasks, sub_ntab=1)
    def apply_async(self, pool, *, kw=dict()):
        '''perform the task asynchronously, i.e. pool.apply_async(...)
        Equivalent to pool.apply_async(self.f, args=self.args, kwds=self.kw)

        kw: {kw}
        '''
        return pool.apply_async(self.f, args=self.args, kwds=self.get_kw(kw=kw))

    # # # UPDATE KWARGS # # #
    def get_kw(self, kw=dict()):
        '''gets self.kw, but updated with values from kw. (doesn't alter self.kw)
        [EFF] if kw is empty, return self.kw instead of making a copy.
        '''
        return {**self.kw, **kw} if kw else self.kw

    # # # DISPLAY # # #
    def _repr_contents(self):
        result = [f'f={self.f.__name__}']
        if len(self.args)>0:
            result.append(f'args={self.args}')
        if len(self.kw)>0:
            result.append(f'kw={self.kw}')
        if self.i is not None:
            result.append(f'i={self.i}')
        return result

    def __repr__(self):
        str_contents = ', '.join(self._repr_contents())
        return f'{self.__class__.__name__}({str_contents})'


''' ----------------------------- unique tasks ----------------------------- '''

class CrashIfCalled():
    '''callable which is not intended to be called. Crashes when called.
    Crashes with error message provided in __init__, if any.
    '''
    def __init__(self, errmsg=None):
        self.errmsg = errmsg

    def __call__(self, *args, **kw):
        if self.errmsg is not None:
            raise ValueError(self.errmsg)
        else:
            raise ValueError('CrashIfCalled was called!')


class UniqueTask(Sentinel, Task):
    '''a unique task. f, args, and kwargs, should be defined at the class level, not inside __init__.'''
    def __init__(self, *args__None, **kw__None):
        pass  # do nothing (instead of super().__init__)

    f = CrashIfCalled('UniqueTask.f was called, but not provided')
    args = ()
    kw = dict()

    def new_with(self, *args, **kw):
        '''return self. If args or kwargs provided, crash with InputError instead.'''
        if args or kw:
            raise InputError('UniqueTask.new_with(...) cannot take any arguments.')
        return self

    @classmethod
    def implied(cls, *args, **kw):
        '''return self. If args or kwargs provided, crash with InputError instead.'''
        if args or kw:
            raise InputError('UniqueTask.implied(...) cannot take any arguments.')
        return cls()


UNSET_TASK = UniqueTask('UNSET_TASK')
UNSET_TASK.f = CrashIfCalled('called UNSET_TASK! This is not supposed to happen. Probably forgot to set a task.')


''' ----------------------------- identity task ----------------------------- '''

def identity(x):
    '''returns x. Provided for convenience with Task API.'''
    return x

class IdentityTask(Task):
    '''task which returns its (single) input. Useful for TaskArray, etc.'''
    def __init__(self, x, i=None):
        super().__init__(identity, x, i=i)


''' ----------------------------- Task Containers ----------------------------- '''

@format_docstring(**_paramdocs_tasks)
class TaskContainer(Container):
    '''a container for multiple tasks; each Task is a function, args, & kwargs.
    Calling self will perform all tasks, returning the result (and updating self.result as well).

    tasks: iterable of Task objects
        Tasks to perform. Expected type depends on subclass, see e.g. TaskList or TaskArray.
    assign_task_idx: {assign_task_idx}
    printable_process_name: {printable_process_name}
    errors_ok: {errors_ok}
    result_missing: {result_missing}
    '''
    task_cls = Task  # class for any Tasks made by this TaskContainer
    tasks = alias('data')

    def __init__(self, tasks, *args_super, assign_task_idx=True, printable_process_name=None,
                 errors_ok=False, result_missing=RESULT_MISSING, **kw_super):
        super().__init__(tasks, *args_super, **kw_super)
        if assign_task_idx:
            self.assign_task_idx()
        self.printable_process_name = printable_process_name
        self.errors_ok = errors_ok
        self.result_missing = result_missing

    def assign_task_idx(self):
        '''assign task.i for tasks in self, based on their positions in self.'''
        for i, task in self.enumerate():
            task.i = i

    # # # PROPERTIES # # #
    @property
    def printable_process_name(self):
        '''return the name to be used for progress updates, if any.
        If None, use the default: "[type(self)].__call__".
        '''
        result = getattr(self, '_printable_process_name', None)
        if result is None:
            result = f'{type(self).__name__}.__call__'
        return result
    @printable_process_name.setter
    def printable_process_name(self, value):
        '''set the name to be used for progress updates.'''
        self._printable_process_name = value

    @property
    def errors_ok(self):
        '''tuple of Exception types which are okay for tasks to raise.
        setting self.errors_ok = False --> use empty tuple, i.e. no errors are okay.
        setting self.errors_ok = errtype --> use errors_ok = (errtype,).
        
        setting errors_ok will crash if it includes any parent class of KeyboardInterrupt,
            e.g. errors_ok=BaseException will crash, but errors_ok=Exception will be fine.

        See also: self.errors_ok_tuple
        '''
        return self._errors_ok
    @errors_ok.setter
    def errors_ok(self, value):
        '''set self.errors_ok to value, converting to tuple if needed. See also: self.errors_ok_tuple'''
        self._errors_ok = self.errors_ok_tuple(value)

    def errors_ok_tuple(self, value=UNSET):
        '''returns tuple of okay errors. UNSET --> self.errors_ok.
        False --> (). errtype --> (errtype,).
        if result includes any parent class of KeyboardInterrupt, raises InputError.
        e.g. errors_ok_tuple(BaseException) will crash, but errors_ok_tuple(Exception) will be fine.
        '''
        if value is UNSET:
            return self.errors_ok
        if value is False:
            value = ()
        elif isinstance(value, type):
            value = (value,)
        value = tuple(value)
        for errtype in value:
            if issubclass(errtype, KeyboardInterrupt):
                raise InputError('errors_ok cannot include KeyboardInterrupt or its parent classes.')
        return value

    # # # CALLING # # #
    def init_result(self):
        '''set self.result = container with similar shape as self, filled with RESULT_MISSING.
        Then, return self.result.
        The idea is that self.result[idx] will correspond to the result of self[idx].
        '''
        self.result = self.new_empty(fill=RESULT_MISSING)
        return self.result

    @format_docstring(**_paramdocs_tasks, sub_ntab=2)
    def __call__(self, *, kw=dict(), idx=None, reset=False, skip_done=False,
                 ncpu=None, timeout=None, ncoarse=1,
                 print_freq=None, errors_ok=UNSET, result_missing=UNSET):
        ''''perform all tasks in self, returning the results.

        OPTIONS (AFFECTS ALL TASKS)
            kw: {kw}

        OPTIONS (AFFECTS WHICH TASKS ARE PERFORMED)
            idx: None or iterable of indices
                None --> perform all tasks in self.
                iterable of indices --> perform only these tasks.
            reset: {reset}
            skip_done: {skip_done}

        OPTIONS (AFFECTS MULTIPROCESSING STRATEGY)
            ncpu: {ncpu}
            timeout: {timeout}
            ncoarse: {ncoarse}

        OPTIONS (MISC)
            print_freq: {print_freq}
            errors_ok: UNSET or {errors_ok}
                UNSET --> use self.errors_ok.
            result_missing: UNSET or {result_missing}
                UNSET --> use self.result_missing.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        if reset or not hasattr(self, 'result'):
            self.init_result()  # creates self.result.
        idx_todo = [i for i, task in self.enumerate(idx) if (not skip_done) or (result[i] is RESULT_MISSING)]
        if len(idx_todo)==0:  # nothing left to calculate!
            return self.result
        if ncpu is None:
            ncpu = mp.cpu_count()
        if ncoarse > 1:
            coarsened = self.coarsen(ncoarse=ncoarse, idx=idx_todo)
            return coarsened(kw=kw, ncpu=ncpu, print_freq=print_freq, timeout=timeout)

        errors_ok = self.errors_ok_tuple(errors_ok)
        if result_missing is UNSET: result_missing = self.result_missing
        _updater = ProgressUpdater(print_freq, wait=True)  # wait=True --> wait print_freq seconds before first print.
        _N_todo = self.size(idx_todo)  # for updates - number of tasks total in self at these idx.
        _idx_done = set()  # helps with debugging - track which tasks were done this time.
        with TimeLimit(timeout):
            if ncpu > 1:  # multiprocessing
                getters = dict()
                with mp.Pool(processes=ncpu) as pool:
                    for _n, (i, task) in enumerate(self.enumerate(idx_todo)):
                        _updater.print(f'assigning task {_n} of {_N_todo}.')
                        getters[i] = task.apply_async(pool, kw=kw)
                    for _n, (i, getter) in enumerate(getters.items()):
                        _updater.print(f'waiting for results from task {_n} of {_N_todo}.')
                        try:
                            self.result[i] = getters[i].get()
                        except errors_ok:
                            self.result[i] = result_missing
                        _idx_done.add(i)
            else:  # single processor
                for _n, (i, task) in enumerate(self.enumerate(idx_todo)):
                    _updater.print(f'performing task {_n} of {_N_todo}.')
                    try:
                        self.result[i] = task(**kw)
                    except errors_ok:
                        self.result[i] = result_missing
                    _idx_done.add(i)
        _updater.finalize(self.printable_process_name, end='\n')
        return self.result
    
    def coarsen(self, ncoarse=5, *, idx=None):
        '''return a TaskPartition containing TaskGroups of size ncoarse.
        Useful for coarsening a TaskContainer for more efficient multiprocessing;
            grouping tasks together can reduce the overhead of multiprocessing,
            while still allowing for parallel processing as the groups are run in parallel.

        if idx is provided, only group the tasks with those indices.
        '''
        return TaskPartition.from_task_container(self, ncoarse=ncoarse, idx=idx)


@format_docstring(**_paramdocs_tasks)
class TaskList(TaskContainer, ContainerOfList):
    '''a container for multiple tasks; each Task is a function, args, & kwargs.
    Calling self will perform all tasks, returning the result (and updating self.result as well).

    tasks: Tasks, iterables, or callables.
        the Tasks to be performed.
        any non-Task input will be used to create a Task.
        E.g. TaskList(f1, (f2, args2), Task(f3, args3, kwargs3))
            --> [Task(f1), Task(f2, args2), Task(f3, args3, kwargs3)]
    from_list: None or iterable
        optionally, can provide all tasks as a single list, instead of as args.
        if provided, *tasks must be empty.
    assign_task_idx: {assign_task_idx}
    printable_process_name: {printable_process_name}
    errors_ok: {errors_ok}
    result_missing: {result_missing}
    '''
    def __init__(self, *tasks, from_list=None, **kw_super):
        if from_list is not None:
            if len(tasks)>0:
                raise InputConflictError('TaskList(...) expects *tasks or from_list, but not both.')
            tasks = from_list
        result = []
        for i, task in enumerate(tasks):
            result.append(self.task_cls.implied(task))
        super().__init__(result, **kw_super)


class TaskArray(TaskContainer, ContainerOfArray):
    '''a container for multiple tasks; each Task is a function, args, & kwargs.
    Calling self will perform all tasks, returning the result (and updating self.result as well).

    tasks: array-like of Tasks, iterables, or callables.
        the Tasks to be performed.
        any non-Task input will be used to create a Task.
        E.g. TaskArray([f1, (f2, args2), Task(f3, args3, kwargs3)])
            --> [Task(f1), Task(f2, args2), Task(f3, args3, kwargs3)]
    shape: None or tuple
        shape that the array of tasks should have.
        if None, infer tasks shape using self.task_nest_shape(tasks).
        if provided, use the specified shape instead of trying to infer it.
    checks: bool
        whether to perform checks on the inputs, and possibly convert inputs if needed.
        if False, assumes that tasks is already a numpy array of Task objects.
    assign_task_idx: {assign_task_idx}
    printable_process_name: {printable_process_name}
    errors_ok: {errors_ok}
    result_missing: {result_missing}
    '''
    def __init__(self, tasks, *, shape=None, checks=True, **kw_super):
        if checks:
            if shape is None:
                shape = self.task_nest_shape(tasks)
            tarr = np.empty(shape, dtype=object)
            for idx in np.ndindex(*shape):  # idx is a tuple of indices.
                # Assign tarr[idx] = tasks[idx], for all idx.
                # But, tasks might not be a numpy array, so we need to do multi-indexing.
                ti = tasks
                if len(idx)==0:
                    ti = ti[()]
                else:
                    for j in idx:
                        ti = ti[j]
                tarr[idx] = ti
            # convert each element of array into a Task
            for idx, task in np.ndenumerate(tarr):
                tarr[idx] = self.task_cls.implied(task)
            tasks = tarr
        super().__init__(tasks, **kw_super)

    @staticmethod
    def task_nest_shape(nested_list):
        '''returns the implied shape for numpy object array of tasks from nested_list.
        This will be the most natural shape to use if each element of nested_list
        is an iterable (e.g. tuple) or callable (e.g. function or Task),
        and there is no desire to iter(f) for any callable f in the nested_list.
        '''
        return nest_shape(nested_list, is_element=is_task_element)

    @classmethod
    def empty(cls, shape):
        '''return a TaskArray of shape shape, filled with UNSET_TASK.'''
        return cls(np.full(shape, UNSET_TASK, dtype=object), checks=False, assign_task_idx=False)


class TaskContainerCallKwargsAttrHaver():
    '''object with attrs corresponding to the main kwargs for TaskContainer.__call__.
    This includes ncpu, timeout, and ncoarse.
    It does not include kw, idx, reset, or skip_done.
    '''
    ncpu = simple_property('_ncpu', setdefaultvia='_default_ncpu',
        doc=_paramdocs_tasks['ncpu'] + '''
        see also: self.get_ncpu() to read actual number of cpus when self.ncpu is None.''')
    def _default_ncpu(self):
        '''returns default for ncpu. Here, returns None; i.e., use multiprocessing.cpu_count().'''
        return None

    def get_ncpu(self):
        '''returns ncpu, but if None, return multiprocessing.cpu_count() instead.
        (This is for convenience; using None will also work with any methods defined here.)
        '''
        return mp.cpu_count() if self.ncpu is None else self.ncpu

    timeout = simple_property('_timeout', setdefaultvia='_default_timeout', doc=_paramdocs_tasks['timeout'])
    def _default_timeout(self):
        '''returns default for timeout. Here, returns None; i.e., no time limit.'''
        return None

    ncoarse = simple_property('_ncoarse', setdefaultvia='_default_ncoarse', doc=_paramdocs_tasks['ncoarse'])
    def _default_ncoarse(self):
        '''returns default for ncoarse. Here, returns 1; i.e., do not group tasks.'''
        return 1

    print_freq = simple_property('_print_freq', setdefaultvia='_default_print_freq', doc=_paramdocs_tasks['print_freq'])
    def _default_print_freq(self):
        '''returns default for print_freq. Here, returns UNSET; i.e., infer from self.verbose.
        (See self.print_freq_explicit to get the actual value of print_freq.)
        '''
        return UNSET

    print_freq_explicit = alias('print_freq',
        doc='''like self.print_freq, but converts UNSET to value based on self.verbose,
        UNSET --> result depends on self.verbose:
            False or <=0 --> -1
            True or (>=1 and <5) --> None
            >=5 --> 0  (i.e. print every progress update)
            if self.verbose doesn't exist --> None
        if result would be None, instead give DEFAULTS.PROGRESS_UPDATES_PRINT_FREQ.''')
    @print_freq_explicit.getter
    def print_freq_explicit(self):
        '''returns the actual value of print_freq, maybe inferred from self.verbose,
        and if None then convert to DEFAULTS.PROGRESS_UPDATES_PRINT_FREQ.
        '''
        result = self.print_freq
        if result is UNSET:  # infer from self.verbose
            if hasattr(self, 'verbose'):
                if (not self.verbose) or (self.verbose <= 0):
                    result = -1
                elif self.verbose >= 5:
                    result = 0
                elif self.verbose:
                    result = None
            else:
                result = None
        if result is None:
            result = DEFAULTS.PROGRESS_UPDATES_PRINT_FREQ
        return result

    def check_pickle(self, x=None):
        '''checks that self (or, x, if provided) is pickleable, by pickling then unpickling.
        Returns result of unpickling. Useful for debugging.
        '''
        obj = self if x is None else x
        dump = pickle.dumps(obj)
        load = pickle.loads(dump)
        return load


''' ----------------------------- TaskGroup (helps with coarsen) ----------------------------- '''

@format_docstring(**_paramdocs_tasks)
class TaskGroup(Task, ContainerOfDict):
    '''treats a dict of Tasks as if it was a single Task.

    Helps with coarsen. Users probably don't need to think about TaskGroup directly.

    tasks: dict of Tasks
    errors_ok: tuple of Exception types
        the exception types for which to return result_missing instead of crashing, in case of error.
    result_missing: {result_missing}
    '''
    args = ()   # behave like a Task with no args
    kw = dict() # behave like a Task with no kwargs

    tasks = alias('data')

    def __init__(self, tasks, *, i=None, errors_ok=(), result_missing=RESULT_MISSING):
        self.tasks = dict(tasks)
        self.i = i
        self.errors_ok = errors_ok
        self.result_missing = result_missing

    def f(self, *args__None, **kw):
        '''perform all tasks in self, returning the results (as a dict)'''
        result = dict()
        errors_ok = self.errors_ok
        for i, task in self.enumerate():
            try:
                result[i] = task(**kw)
            except errors_ok:
                result[i] = self.result_missing
        return result


@format_docstring(**_paramdocs_tasks)
class TaskPartition(TaskList):
    '''a container for a list of TaskGroups, each with Tasks from the same TaskContainer.
    Knows to call each TaskGroup in parallel,
        but call in series the individual tasks within each TaskGroup.

    Helps with TaskContainer.coarsen.

    assign_task_idx: {assign_task_idx}
    printable_process_name: {printable_process_name}
    '''
    group_cls = TaskGroup  # class for any TaskGroups made by this TaskPartition
    groups = alias('tasks')

    def __init__(self, original, groups, **kw_super):
        self.original = original
        super().__init__(from_list=list(groups), **kw_super)

    @classmethod
    @format_docstring(**_paramdocs_tasks, sub_ntab=1)
    def from_task_container(cls, original, ncoarse=5, *, idx=None, errors_ok=UNSET, result_missing=UNSET):
        '''return a TaskPartition containing TaskGroups of size ncoarse.
        Useful for coarsening a TaskContainer for more efficient multiprocessing;
            grouping tasks together can reduce the overhead of multiprocessing,
            while still allowing for parallel processing as the groups are run in parallel.

        if idx is provided, only group the tasks with those indices.

        errors_ok: UNSET or {errors_ok}
            UNSET --> use original.errors_ok.
        result_missing: {result_missing}
            UNSET --> use original.result_missing.
        '''
        idxs0 = [i for i, task in original.enumerate(idx=idx)]
        idxs = [idxs0[i:i+ncoarse] for i in range(0, len(idxs0), ncoarse)]
        task_mappings = [{i:task for i,task in original.enumerate(idx=jdx)} for jdx in idxs]
        if errors_ok is UNSET: errors_ok = original.errors_ok
        if result_missing is UNSET: result_missing = original.result_missing
        kw_group = dict(errors_ok=errors_ok, result_missing=result_missing)
        printable_process_name = f'{cls.__name__}({original.printable_process_name}, ncoarse={ncoarse})'
        return cls(original, [cls.group_cls(tasks, **kw_group) for tasks in task_mappings],
                   printable_process_name=printable_process_name)

    def _assign_result_i(self, result_i, i):
        '''assign result_i to self.result[i], and also self.original.result[j] for all relevant j.'''
        self.result[i] = result_i
        for j, task in self[i].enumerate():
            self.original._assign_result_i(result_i[j], j)

    @format_docstring(**_paramdocs_tasks, sub_ntab=1)
    def __call__(self, *, kw=dict(), ncpu=None, timeout=None, print_freq=None):
        '''perform all tasks in all groups of self. return results formatted to original shape.

        kw: {kw}
        ncpu: {ncpu}
        timeout: {timeout}

        The following kwargs are acceptable for TaskContainer but NOT here:
            idx, reset, skip_done, ncoarse, errors_ok, result_missing.
        That is because idx, reset, skip_done, and ncoarse are not relevant,
        while errors_ok and result_missing are handled by the TaskGroups.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        call_result = super().__call__(kw=kw, ncpu=ncpu, timeout=timeout, print_freq=print_freq,
                                       errors_ok=False, # result_missing doesn't matter because errors_ok=False
                                       )
        for i, group in self.enumerate():
            result_i = call_result[i]
            for j, task in group.enumerate():
                self.original.result[j] = result_i[j]
        return self.original.result


''' ----------------------------- Test Functions ----------------------------- '''

# test functions to help test/develop/debug multiprocessing codes^
def mptest_add100(x):
    '''adds 100 to x'''
    return x + 100

def mptest_sleep(t):
    '''sleeps for t seconds, then returns t'''
    time.sleep(t)
    return t

def mptest_sleep_add100(t):
    '''sleeps for t seconds, then returns t+100'''
    time.sleep(t)
    return t + 100

def mptest_echo(*args, **kw):
    '''print inputs then return (args, kw)'''
    print('args:', args)
    print('kw:', kw)
    return args, kw

class MPTestClass():
    '''empty class useful for multiprocessing testing.'''
    mem = dict()

    # def __init__(self):
    #     self.mem = dict()

    def update_mem(self, key, value):
        '''update self.mem with key, value. Print mem after.'''
        self.mem[key] = value
        print('after update_mem:', self.mem)

    def get_mem(self, key):
        '''return self.mem[key]. print mem before.'''
        print('before get_mem:', self.mem)
        return self.mem[key]



"""
File Purpose: Line: 1D lines; movie of lines for 2D data.
[TODO] why does plt.legend() move around during animation (if loc not specified)?
"""
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from .labels import xarray_title_plot_node
from .movies import MoviePlotElement, MoviePlotNode, MovieOrganizerNode
from .plot_settings import PlotSettings
from .plot_tools import (
    set_margins_if_provided, set_lims_if_provided,
    infer_movie_dim, infer_xy_dims, infer_xyt_dims,
    calculate_vlims, calculate_lims_from_margin,
) 
from ..errors import PlottingAmbiguityError, DimensionError
from ..tools import (
    alias, alias_key_of, alias_child, simple_property,
    UNSET,
    pcAccessor, take_along_dimensions, xarray_nondim_coords,
    xarray_object_coords_to_str,
    xarray_as_array, xarray_fill_coords, xarray_ensure_dims,
)
from ..defaults import DEFAULTS


@PlotSettings.format_docstring()
class XarrayLinePlotElement(MoviePlotElement):
    '''plot line of 1D data.
    
    array: xarray.DataArray, probably ndim=1.
        the data to be plotted.
        coords with ndim>0, dtype=object will be converted to strings suitable for ticklabels.
    label: {label}
    '''
    def __init__(self, array, *, ax=None,
                 aspect=None,  # <-- different from PlotSettings default!
                 **kw_super):
        super().__init__(aspect=aspect, **kw_super)
        self._ax_init = ax
        if array.ndim != 1:
            raise DimensionError(f'line expected array.ndim=1, but got ndim={array.ndim}.')
        self.array = xarray_object_coords_to_str(array, maxlen=DEFAULTS.PLOT.XTICKLABEL_MAXLEN, ndim_min=1)
        self.init_line()

    array = alias_key_of('data', 'array', doc='''array for plot. Internally, stored at self.data['array']''')
    line = simple_property('_line', doc='''line plot (instance of matplotlib.lines.Line2D)''')
    ax = alias_child('line', 'axes', if_no_child=None, doc='''the axes containing this line plot''')
    fig = alias_child('ax', 'figure', if_no_child=None, doc='''figure containing this line plot''')

    # # # PLOTTING THIS ELEMENT # # #
    def init_line(self):
        '''initialize the line plot; actually plot the data.
        stores plotted object in self.line and returns self.line.
        '''
        array = self.array
        # plot settings / kwargs / bookkeeping
        kw_plot = self.plot_settings.get_mpl_kwargs('xarray.DataArray.plot.line')
        if 'label' in kw_plot:
            kw_plot['label'] = kw_plot['label'].format(**xarray_nondim_coords(array))
        # make plot
        plotter = array.plot.line
        lines = plotter(ax=self._ax_init, **kw_plot)
        self.line = lines[0]
        if len(lines) > 1:
            raise PlottingAmbiguityError(f'xarray.DataArray.plot.line made {len(lines)} lines, expected 1.')
        # formatting
        aspect = self.plot_settings['aspect']  # xarray plotter doesn't like 'aspect' kwarg; assign here instead.
        if aspect is not None:
            self.ax.set_aspect(aspect)
        set_margins_if_provided(self.plot_settings, ax=self.ax)
        set_lims_if_provided(self.plot_settings, ax=self.ax)
        return self.line

    # # # UPDATING (REQUIRED BY PARENT) # # #
    def update_data(self, data):
        '''update the plot using data['array'].
        return the list of all updated matplotlib Artist objects (i.e., [self.line])
        '''
        array = data['array']
        line = self.line
        linex = line.get_xdata()
        line.set_data(linex, array)
        self.array = array  # <-- bookkeeping
        return [self.line]


@pcAccessor.register('line', totype='array')
@PlotSettings.format_docstring()
class XarrayLine(MoviePlotNode):  # [TODO] refactor PlotDimsMixin to use it here.
    '''MoviePlotNode of line.
    stores an XarrayLinePlotElement & has methods for plotting & updating the plot.
    "line" refers to a matplotlib.lines.Line2D object, e.g. the result of plt.plot.

    array: xarray.DataArray, probably ndim=2.
        the data to be plotted. if ndim=1, can still plot, but nothing to animate.
    t: None or str
        the array dimension which frames will index. E.g. 'time'.
        None -> infer from array & any other provided dimensions.
    x: None or str
        if provided, tells dimensions for the x plot axes.
        None -> infer from array & any other provided dimensions.
    ax: None or Axes
        the attached mpl.axes.Axes object.
        None --> will use self.ax=plt.gca() when getting self.ax for the first time.

    init_plot: {init_plot}

    label: {label}

    {kw_ax_margin}
    aspect: {aspect}

    title: {title}
    title_font: {title_font}
    title_y: {title_y}
    title_kw: {title_kw}
    '''
    element_cls = XarrayLinePlotElement

    def __init__(self, array, t=None, *, x=None, ax=None,
                 init_plot=PlotSettings.default('init_plot'),  # <- could go in kw_super, but explicit is nice.
                 label=PlotSettings.default('label'),
                 aspect=None,  # <-- different from PlotSettings default!
                 **kw_super):
        array = xarray_fill_coords(array, need=[x, t])
        array = xarray_ensure_dims(array, ({x, t} - {None}), promote_dims_if_needed=True)
        self.array = array
        self._ax_init = ax
        # infer dims ([TODO]: refactor plot dims code & offload this logic to PlotDimsMixin.)
        if array.ndim != 1:
            t = infer_movie_dim(array.dims, t, fail_ok=True)
        if (array.ndim != 1) and (t is None):
            errmsg = 't=None but is required (when array.ndim != 2). But cannot infer t from data provided.'
            raise PlottingAmbiguityError(errmsg)
        if x is None:
            x = infer_xy_dims(array.dims, x=x, y=None, exclude=[t], fail_ok=True)[0]
        if x is None:
            errmsg = 'x=None but is required. But cannot infer x from data provided.'
            raise PlottingAmbiguityError(errmsg)
        self.t = t
        self.x = x
        # super init
        kw_super.update(init_plot=init_plot, label=label, aspect=aspect)
        super().__init__(**kw_super)

    # # # PROPERTIES # # #
    t_plot_dim = alias('t')

    ax = alias_child('obj', 'ax', if_special_child={None: None, UNSET: None},
        doc='''mpl.axes.Axes where this XarrayLine is plotted, or None if not plotted.''')

    fig = alias_child('obj', 'fig', if_special_child={None: None, UNSET: None},
        doc='''figure where this XarrayLine is plotted, or None if not plotted.''')

    # # # PLOTTING METHODS (REQUIRED BY PARENT CLASS) # # #
    def init_plot(self):
        '''plot for the first time. Save the ContourPlotElement at self.obj'''
        self._init_plot_checks()
        frame = self.plot_settings['init_plot_frame']
        data = self.get_data_at_frame(frame)
        # get settings for plot
        kw_plot = self.plot_settings.get_mpl_kwargs('pc.XarrayLinePlotElement')
        # -- determine ylims (based on the entire array, not just this frame)
        d = kw_plot
        robust = d.get('robust', False)
        ylim = d.pop('ylim', None)
        ymargin = 0 if (ylim is not None) else d.get('ymargin', None)
        ymin, ymax = (None, None) if (ylim is None) else ylim
        ymin, ymax = calculate_vlims(self.array, vmin=ymin, vmax=ymax, robust=robust)
        ymin, ymax = calculate_lims_from_margin(ymin, ymax, margin=ymargin, x='y')
        d['ylim'] = (ymin, ymax)
        # >> actually plot the contour <<
        self.obj = self.element_cls(data['array'], ax=self._ax_init, **kw_plot)
        # -- if add_labels, add title:
        if self.plot_settings['add_labels']:
            self.plot_title()
        # bookkeeping
        self.frame = frame

    # [TODO] encapsulate similarities between this & XarrayImage???
    def get_data_at_frame(self, frame):
        '''returns {'array': array at this frame}, properly transposed & ready for plotting.'''
        t = self.t_plot_dim
        result = self.array
        if t is not None:  # there is a dimension for the time axis.
            result = result.isel({t: frame})
        return {'array': result}

    def get_nframes_here(self):
        '''return the number of frames that could be in the movie, based on this node.'''
        t = self.t_plot_dim
        if t is None:
            return 0
        return len(self.array.coords[t])

    # # # TITLE # # #
    def plot_title(self):
        '''adds title (as a MovieTextNode) on self.ax.
        raise PlottingAmbiguityError if title already plotted
            (this prevents having multiple title nodes).
        '''
        if hasattr(self, 'title_node'):
            raise PlottingAmbiguityError(f'{type(self).__name__} title was already plotted!')
        title = self.plot_settings['title']
        if title is None:
            plt.title('')  # remove xarray's default title.
        else:
            kw_title = {**self.plot_settings.kw, 'init_plot': True}
            title_node = xarray_title_plot_node(self.array, title, t=self.t_plot_dim,
                                                ax=self.ax, parent=self, **kw_title)
            self.title_node = title_node  # <-- for bookkeeping/debugging



''' --------------------- MULTIPLE LINES --------------------- '''

@pcAccessor.register('lines')
class XarrayLines(MovieOrganizerNode):
    '''MovieOrganizerNode for organizing multiple XarrayLines.

    array: xarray.DataArray or xarray.Dataset
        the array to plot lines from. Any number of dimensions is allowed.
        if dataset, will be converted to array using to_array(dim='variable').
        Note that dims other than t & x should be "not too long" otherwise plot will have lots of lines.
    dims: None , str, or list of str
        dimensions to iterate over; plot one line per point in these dimensions.
        E.g. dims = ['fluid', 'component'] --> plot one line per fluid-component pair.
        [TODO] improve formatting; currently uses default cycler for line properties,
            but if using multiple dims it would be nicer to have a different cycler for each dim,
            e.g. colors for fluid, linestyles for component.
        None --> infer from array.dims, t, and x.
    t: None or str
        dimension name for the time axis (for movies).
        None --> infer from the array coords. See also: DEFAULTS.PLOT.DIMS_INFER
        If no time dimension (None provided & can't infer), that's okay, but animation will fail.
    x: None or str
        dimension name for the x axis.
        None --> infer from the array coords. See also: DEFAULTS.PLOT.DIMS_INFER
    cstyles: None or dict of {{coordname: dict or list of tuples of (val, dict of kwargs for XarrayLine)}}
        if provided, pass these dicts to individual lines with corresponding scalar val for coord.
        use tuples of values to test equality instead of indexing a dict.
        E.g., styles={{'fluid': [('e', dict(ls='--')), ('H_II', dict(color='blue'))]}}
            would ensure dashed line when arr['fluid']=='e', blue line when arr['fluid']=='H_II',
            and have no effect whenever arr['fluid'] is not scalar, doesn't exist, or isn't 'e' or 'H_II'.
    cstyles_default: bool
        tells how to handle conflict between kwargs from cstyles and kwargs passed directly to self.
        True --> treat cstyles as 'defaults'; kwargs from self override kwargs from cstyles.
        False --> kwargs from cstyles take precedence.

    robust: {robust}
    ymargin: {ymargin}
    add_legend: bool
        whether to plot a legend, by default.
    label: {label}
        UNSET --> 'dim=val' for each line, for each dim in dims, e.g. 'fluid=e-, component=x'
    '''
    def __init__(self, array, dims=None, t=None, *, x=None, ax=None, add_legend=True,
                 label=UNSET, cstyles=None, cstyles_default=False, **kw):
        name = '' if (getattr(array, 'name', None) is None) else array.name
        self._dims_init = dims
        self.cstyles = cstyles
        dims_provided = None if dims is None else [dims] if isinstance(dims, str) else dims
        dims_need = [] if dims_provided is None else dims_provided
        if isinstance(array, xr.Dataset):
            array = xarray_as_array(array)
        array = xarray_fill_coords(array, need=[x, t, *dims_need])
        array = xarray_ensure_dims(array, ({x, t, *dims_need} - {None}), promote_dims_if_needed=True)
        self.array = array
        # infer plot dims & get xarray.DataArray for each line
        x, y, t = infer_xyt_dims(self.array.dims, t=t, x=x, exclude=dims_need,
                                 xy_fail_ok=True, t_fail_ok=True)  # [TODO] improve plot_dims code?
        self.x, self.t = x, t  # <-- bookkeeping, store for convenience.
        if x is None:  # note: it's okay if t is None; that just means "single frame, not a movie".
            errmsg = 'Cannot infer x. Provide additional info (x, t, and/or dims) then try again.'
            raise PlottingAmbiguityError(errmsg)
        if dims is None:
            dims = [dim for dim in self.array.dims if dim not in [t, x]]
        else:
            dims = dims_provided
        # labels (for legend)
        if label is UNSET:
            label = ', '.join(['dim={dim}'.replace('dim', dim) for dim in dims])
            if label=='': label = UNSET
        kw['label'] = label
        # list of xarray DataArrays.
        arrs = take_along_dimensions(self.array, dims, atleast_1d=True)
        self.arrs = arrs
        if arrs.ndim >= 2:
            raise NotImplementedError(f"[TODO] XarrayLines across 2D+ ({dims!r}). Workaround: use different x/t/dims.")
        # handle ylims, taking all arrays into account.
        init_plot_settings = PlotSettings(**kw)
        robust = init_plot_settings.get('robust', last_resort_default=False)
        ylim = init_plot_settings.get('ylim', last_resort_default=None)
        ymargin = 0 if (ylim is not None) else init_plot_settings.get('ymargin', last_resort_default=None)
        ymin, ymax = (None, None) if (ylim is None) else ylim
        ymin, ymax = calculate_vlims(self.array, vmin=ymin, vmax=ymax, robust=robust)
        ymin, ymax = calculate_lims_from_margin(ymin, ymax, margin=ymargin, x='y')
        kw['ylim'] = (ymin, ymax)
        # create array of XarrayLine nodes (but don't init_plot yet!)
        kw_line = kw.copy()
        kw_line['title'] = None  # <-- allow self to determine title, not any of the lines.
        lines = np.full(arrs.shape, None, dtype=object)
        for idx, arr in np.ndenumerate(arrs):
            # cstyle management ([TODO] encapsulate this and put it elsewhere...?)
            cstyle_here = dict()
            if cstyles is not None:
                for cname, csty in cstyles.items():
                    if cname in arr.coords and arr[cname].size==1:
                        if isinstance(csty, dict):
                            cstyle_here.update(csty.get(arr[cname].item(), {}))
                        else:  # csty must be an iterable of 2-tuples
                            for cval, cvalstyle in csty:
                                if arr[cname].item() == cval:
                                    cstyle_here.update(cvalstyle)
            kw_here = {**cstyle_here, **kw_line} if cstyles_default else {**kw_line, **cstyle_here}
            # >> actually make the XarrayLine. Don't plot it yet though. <<
            line = XarrayLine(arr, t=t, x=x, ax=ax, init_plot=False, **kw_here)
            lines[idx] = line
            # never mess with title or labels after the first one.
            kw_line['add_labels'] = False
        self.nodes = lines
        # init super
        super().__init__(name=name, **kw)

        if self.plot_settings['init_plot']:
            self.init_plot()

        self.add_legend = add_legend
        if add_legend:
            self.plot_legend()

    def init_plot(self):
        '''plot for the first time: call init_plot on all nodes, and connect to tree.

        This is fundamentally different from MoviePlotNode.init_plot's usual purpose,
            since this is about calling init_plot on the nodes, not on self.obj.
            [TODO] should this function be renamed, for clarity?
        '''
        # checks: skip the usual checks; init_plot on nodes, not self.obj.
        # but, do check if self.init_plot was already called; don't want to call it twice!
        if getattr(self, '_called_init_plot', False):
            errmsg = (f'init_plot was already called; cannot call it again. '
                      f'For {type(self).__name__}(name={self.name!r}) with id={hex(id(self))}')
            raise PlottingAmbiguityError(errmsg)
        for i, line in np.ndenumerate(self.nodes):
            line.init_plot()
            line.parent = self
        # -- if add_labels, add title:
        if self.plot_settings['add_labels']:
            self.plot_title()
        self._called_init_plot = True

    # # # PROPERTIES # # #
    lines = alias('nodes')
    line0 = property(lambda self: self.lines[0], doc='''first line; lines[0].''')
    ax = alias_child('line0', 'ax', doc='''mpl.axes.Axes where the lines are plotted.''')
    fig = alias_child('line0', 'fig', doc='''figure where the lines are plotted.''')

    # # # LEGEND # # #
    def plot_legend(self):
        '''plot a legend! Plots to the right of the axes.
        For more precise control of legend,
            use add_legend=False during self.__init__,
            and just call plt.legend() separately...

        Also fig.set_layout_engine('tight') if layout engine was None;
            otherwise the legend might get cut off at the edges.
        '''
        if self.fig.get_layout_engine() is None:
            self.fig.set_layout_engine('tight')
        self.ax.legend(loc='upper left', bbox_to_anchor=(1.02, 0.95))

    # # # TITLE # # #
    def plot_title(self):
        '''adds title (as a MovieTextNode) on self.ax.
        raise PlottingAmbiguityError if title already plotted
            (this prevents having multiple title nodes).
        '''
        if hasattr(self, 'title_node'):
            raise PlottingAmbiguityError(f'{type(self).__name__} title was already plotted!')
        title = self.plot_settings['title']
        if title is None:
            plt.title('')  # remove xarray's default title.
        else:
            kw_title = {**self.plot_settings.kw, 'init_plot': True}
            title_node = xarray_title_plot_node(self.array, title, t=self.t,
                                                ax=self.ax, parent=self, **kw_title)
            self.title_node = title_node  # <-- for bookkeeping/debugging



"""
File Purpose: base quantities for MuramCalculator
"""

from ...defaults import DEFAULTS
from ...dimensions import SINGLE_FLUID
from ...errors import (
    FluidValueError, FormulaMissingError, LoadingNotImplementedError,
    InputError, InputMissingError,
)
from ...mhd import MhdBasesLoader
from ...tools import (
    simple_property,
    UNSET,
)


''' --------------------- MuramBasesLoader --------------------- '''

class MuramBasesLoader(MhdBasesLoader):
    '''base quantities based on Muram output.'''
    @known_var(dims=['snap'])
    def get_r(self):
        '''mass density. (directly from Muram)
        assumes single fluid mode, i.e. result corresponds to the single fluid from Muram.
        '''
        self.assert_single_fluid_mode('r')
        return self.load_maindims_var_across_dims('result_prim_r', u='mass_density', dims=['snap'])

    @known_var(dims=['snap', 'component'])
    def get_u(self):
        '''velocity. (directly from Muram)
        assumes single fluid mode, i.e. result corresponds to the single fluid from Muram.
        '''
        self.assert_single_fluid_mode('u')
        return self.load_maindims_var_across_dims('result_prim_u', u='speed', dims=['snap', 'component'])

    @known_var(dims=['snap'])
    def get_e(self):
        '''energy density. (directly from Muram)
        Per unit volume, e.g. the SI units would be Joules / meter^3.
        assumes single fluid mode, i.e. result corresponds to the single fluid from Muram.
        '''
        self.assert_single_fluid_mode('e')
        return self.load_maindims_var_across_dims('result_prim_e', u='energy_density', dims=['snap'])

    @known_var(dims=['snap', 'component'])
    def get_B(self):
        '''magnetic field. (directly from Muram)'''
        return self.load_maindims_var_across_dims('result_prim_b', u='b', dims=['snap', 'component'])

    @known_var
    def get_gamma(self):
        '''adiabatic index. Assumed to be 5/3.'''
        return 5/3.



"""
File Purpose: misc tests for eppic hookup in PlasmaCalcs.
"""

import os
import pickle
import pytest

import PlasmaCalcs as pc
pc.DEFAULTS.pic_ambiguous_unit = dict(u_t=1)  # must be defined before loading an EppicCalculator.

HERE = os.path.dirname(__file__)


''' --------------------- basics --------------------- '''

def get_eppic_calculator(eppic_i='test_eppic_basics.i', **kw_init):
    '''make an eppic calculator'''
    with pc.InDir(HERE):
        ec = pc.EppicCalculator.from_here(eppic_i, **kw_init)
    return ec


''' --------------------- help --------------------- '''

def test_eppic_calculator_help():
    '''ensure that eppic calculator help doesn't crash.'''
    print('\n\n')
    print('--------------------- [test_eppic_calculator_help] START ---------------------')
    print('\n\n')
    # as classmethod
    pc.EppicCalculator.cls_help()  # allow to print this; it's not huge output
    pc.EppicCalculator.cls_help(print=False)
    pc.EppicCalculator.cls_help('', print=False)
    pc.EppicCalculator.cls_help('deltafrac_n', print=True)  # allow to print this; it's not huge output
    pc.EppicCalculator.cls_help('mod', print=False)
    pc.EppicCalculator.cls_help('q', print=False)
    pc.EppicCalculator.cls_help('', modules=True, print=False)
    pc.EppicCalculator.cls_help('', modules=True, dense=False, print=False)
    pc.EppicCalculator.cls_help('1+2', print=False)
    pc.EppicCalculator.cls_help('1+(2+3)', print=False)
    pc.EppicCalculator.cls_help('var_that_isnt_defined_anywhere')  # allow to print this; it's not huge output
    # as instance method
    ec = get_eppic_calculator()
    ec.help(print=False)  # could print this but decided not to.
    ec.help('', print=False)
    ec.help('deltafrac_n', print=True)  # allow to print this; it's not huge output
    ec.help('mod', print=False)
    ec.help('q', print=False)
    ec.help('', modules=True, print=False)
    ec.help('', modules=True, dense=False, print=False)
    ec.help('1+2', print=False)
    ec.help('1+(2+3)', print=False)
    ec.help('var_that_isnt_defined_anywhere')    # allow to print this; it's not huge output

    # calling help as classmethod by accident should crash:
    with pytest.raises(pc.InputError):
        pc.EppicCalculator.help()
    # calling class help for var depending on instance values should crash:
    with pytest.raises(pc.InputError):
        pc.EppicCalculator.cls_help('nusj')  # nusj needs to know ec.fluid...
    # calling help from instance should not crash:
    ec.help('nusj', print=False)
    # even if fluids change, still shouldn't crash:
    ec.fluid = 0
    ec.help('nusj', print=False)
    ec.fluid = [0,1]
    ec.help('nusj', print=False)
    print('\n\n')
    print('--------------------- [test_eppic_calculator_help] FINISH ---------------------')
    print('\n\n')


''' --------------------- units --------------------- '''

def test_eppic_coords_units():
    '''ensure coords_units works as intended.
    See also: issue #4
    '''
    ec = get_eppic_calculator()
    assert ec.units == 'si'
    assert ec.coords_units is None  # i.e. "always match ec.units"
    assert ec.coords_units_explicit == 'si'
    ec.units = 'raw'
    assert ec.coords_units_explicit == 'raw'  # matches ec.units
    # ensure coords_units still matches ec.units even ec.using(...)
    ec.units = 'si'
    with ec.using(coords_units='raw'):
        pass
    assert ec.coords_units_explicit == 'si'
    ec.units = 'raw'
    assert ec.coords_units_explicit == 'raw'  # still matches ec.units
    ec.coords_units = 'raw'
    ec.units = 'si'
    assert ec.coords_units == 'raw'
    assert ec.coords_units_explicit == 'raw'  # no longer matches ec.units


''' --------------------- timescales --------------------- '''

def test_eppic_timescales():
    '''ensure the EppicCalculator.timescales() method is working properly.'''
    # [TODO] actually test the values. (currently, just testing that it doesn't crash.)
    ec = get_eppic_calculator()
    ec.set_vars_from_inputs()
    ec.timescales()

def test_eppic_choose_params():
    '''ensure the EppicCalculator.choose_params() method is working properly.'''
    # [TODO] actually test the values... (currently, just testing that it doesn't crash.)
    ec = get_eppic_calculator()
    ec.set_vars_from_inputs()
    ec.choose_params()


''' --------------------- fft_keep --------------------- '''

def test_fft_keep():
    '''ensure that the fft_keep property works as intended.'''
    ec = get_eppic_calculator()
    ec.fft_keep = None
    slices0 = ec.fft_slices.copy()
    ec.fft_keep = 0.1
    slices1 = ec.fft_slices.copy()
    assert slices1 != slices0  # setting keep adjusts fft_slices.
    with ec.using(fft_keep=None):
        assert ec.fft_slices == slices0  # setting keep adjusts fft_slices.
    # exiting a "using" block resets keep to old value, adjusting fft_slices.
    assert ec.fft_slices == slices1
    ec.fft_keep = 0.2
    for x in ec.maindims:
        assert ec.fft_slices.get_slice(f'freq_{x}') == slice(0.4, 0.6, None)
    ec.fft_keep = 0.3
    for x in ec.maindims:
        assert ec.fft_slices.get_slice(f'freq_{x}') == slice(0.35, 0.65, None)
    assert len(ec.maindims)>0  # i.e., that loop iterated over a non-empty list^
    # test that using half works too.
    x = ec.maindims[0]
    ec.fft_keep = 0.3
    ec.fft_half = x
    assert ec.fft_slices.get_slice(f'freq_{x}') == slice(0.5, 0.65, None)
    for y in ec.maindims[1:]:
        assert ec.fft_slices.get_slice(f'freq_{y}') == slice(0.35, 0.65, None)


''' --------------------- multiprocessing --------------------- '''

def test_eppic_pickling():
    '''ensure that EppicCalculator can be pickled & unpickled.'''
    ec = get_eppic_calculator()
    # immediately after initializing:
    dump = pickle.dumps(ec)
    _tmp = pickle.loads(dump)
    # after running a calculation (which doesn't depend on any vars):
    ec('0')
    dump = pickle.dumps(ec)
    _tmp = pickle.loads(dump)
    # after setting vars:
    ec.set_vars_from_inputs()
    dump = pickle.dumps(ec)
    _tmp = pickle.loads(dump)
    # after getting a set var
    ec('n')
    dump = pickle.dumps(ec)
    _tmp = pickle.loads(dump)

def test_eppic_simple_mp_compatible():
    '''ensure that multiprocessing doesn't cause crashes when doing simple things.
    Not guaranteed to test that multiprocessing is actually occurring...
    but guaranteed to ensure that multiprocessing isn't interfering with simple things.
    '''
    ec = get_eppic_calculator()
    ec.ncpu = 2
    # immediately after initializing
    ec('0')
    # from input deck
    ec('n0')  # should get n0 for multiple fluids
    # when setting vars
    ec.set_vars_from_inputs()
    # after setting vars
    ec('n')



"""
File Purpose: handle unit conversions
"""

from ..errors import (
    UnitsError, UnitsUnknownError,
    InputConflictError,
)
from ..tools import (
    alias, simple_property,
    format_docstring, UNSET,
)
from ..defaults import DEFAULTS


''' --------------------- CONSTANTS & KNOWN SYMBOLS --------------------- '''

# DIMENSIONLESS
DIMENSIONLESS = '1'  # used to indicate dimensionless quantity.

# SI BASE UNIT SYMBOLS
SI_BASE_UNIT_SYMBOLS = (
    "kg",   # M; mass; kilograms
    "m",    # l; length; meters
    "s",    # t; time; seconds
    "A",    # A; current; amperes
    "K",    # K; temperature; kelvin
    )

# OTHER SI UNITS
#    [TODO] use intuitive definitions instead of SI base units? e.g. J = N m
#    The intuitive definitions might help to ensure there are no typos;
#    e.g. it's easier to check {'S': 'Ohm-1'} than {'S': 'kg-1 m-2 s3 A2'}
SI_OTHER_UNIT_SYMBOLS = {
    'Hz'     : 's-1',               # hertz
    'N'      : 'kg m s-2',          # newton
    'Pa'     : 'kg m-1 s-2',        # pascal
    'J'      : 'kg m2 s-2',         # joule
    'W'      : 'kg m2 s-3',         # watts
    'C'      : 's A',               # coulomb
    'V'      : 'kg m2 s-3 A-1',     # volt
    'F'      : 'kg-1 m-2 s4 A2',    # farad
    'Ohm'    : 'kg m2 s-3 A-2',     # ohm
    'S'      : 'kg-1 m-2 s3 A2',    # siemens   # inverse ohm
    'Wb'     : 'kg m2 s-2 A-1',     # weber   # magnetic flux
    'T'      : 'kg s-2 A-1',        # tesla
    'H'      : 'kg m2 s-2 A-2',     # ???   # permeability = H / m
}

# KNOWN CONVERSION SYMBOLS
KNOWN_CONVERSION_SYMBOLS = {
    'dimensionless'   : '1',        #'1': 'dimensionless',
    'time'            : 's',       't': 'time',
    'length'          : 'm',       'l': 'length',
    'mass'            : 'kg',      'M': 'mass',
    'current'         : 'A',
    'temperature'     : 'K',       'Th': 'temperature', 'tg': 'temperature',
    'charge'          : 'C',       'q': 'charge',
    'frequency'       : 'Hz',
    'speed'           : 'm s-1',   'u': 'speed',
    'acceleration'    : 'm s-2',   'a': 'acceleration',
    'area'            : 'm2',
    'volume'          : 'm3',
    'wave_number'     : 'm-1',     'lambda': 'wave_number',
    'momentum'        : 'M u',     'p': 'momentum',
    'force'           : 'N',       'f': 'force',
    'pressure'        : 'Pa',      'P': 'pressure', 'energy_density': 'pressure',
    'energy'          : 'J',       'E': 'energy',  'kT': 'energy',
    'power'           : 'W',
    'number_density'  : 'm-3',     'n': 'number_density', 'nr': 'number_density',
    'mass_density'    : 'n M',     'r': 'mass_density', 'rho': 'mass_density',
    'charge_density'  : 'n C',     'nq': 'charge_density', 'rhoc': 'charge_density',
    'current_density' : 'A m-2',   'j': 'current_density',
    'energy_density'  : 'J m-3',
    'flux'            : 'n u',
    'momentum_density': 'n M u',  'pr': 'momentum_density',
    'permittivity'    : 'F m-1',
    'permeability'    : 'H m-1',
    'potential'       : 'V',       'phi': 'potential',
    'capacitance'     : 'F',
    'resistance'      : 'Ohm',
    'conductance'     : 'S',
    'b_flux'          : 'Wb',
    'b_field'         : 'T',       'b': 'b_field',
    'h_field'         : 'A m-1',
    'inductance'      : 'H',
    'e_field'         : 'V m-1',   'ef': 'e_field',
    'opacity'         : 'l2 m-2',
}

# dict of all symbols which we can convert into SI base units
KNOWN_CONVERSIONS = {DIMENSIONLESS: DIMENSIONLESS,
                    **{key: key for key in SI_BASE_UNIT_SYMBOLS},
                    **SI_OTHER_UNIT_SYMBOLS,
                    **KNOWN_CONVERSION_SYMBOLS}

# PHYSICAL_CONSTANTS_SI = dict of {constant: (value, units string)}
PHYSICAL_CONSTANTS_USTRS = {
    'amu'     : 'M',            # atomic mass unit
    'c'       : 'u',            # speed of light
    'kB'      : 'J K-1',        # boltzmann constant
    'eps0'    : 'permittivity', # permittivity of free space
    'mu0'     : 'permeability', # permeability of free space
    'qe'      : 'q',            # elementary charge
    'me'      : 'M',            # electron mass
    'qme'     : 'q M-1',        # q / m_e
    'hplanck' : 'J s',          # planck constant (not hbar)
    'm_proton': 'M',            # proton mass
    'eV'      : 'J',            # electron volt
    'eV kB-1' : 'K',            # eV / kB
    'me amu-1': '1',            # m_e / amu
}
for _key, _alias in (('qe', 'q_e'), ('me', 'm_e'), ('m_proton', 'm_p'), ('hplanck', 'h')):
    PHYSICAL_CONSTANTS_USTRS[_alias] = PHYSICAL_CONSTANTS_USTRS[_key]
del _key, _alias

PHYSICAL_CONSTANTS_SI = {cname: (DEFAULTS.PHYSICAL.CONSTANTS_SI[cname], ustr) \
    for cname, ustr in PHYSICAL_CONSTANTS_USTRS.items()}
# [TODO] allow physical constants to be involved in the conversion factor notation, in general.


''' --------------------- helper functions --------------------- '''

# FUNCTIONS WHICH READ AND MODIFY THE INPUT STRINGS
def _get_symbol_and_exponent(unit_string):
    '''
    Extract the unit symbol and the exponent from the unit string
    :param unit_string: string of units (e.g. "m-1")
    :return: unit symbol, exponent (e.g. "m", -1)
    '''
    symbol = unit_string
    exponent = 1
    for position in range(len(unit_string)):
        letter = unit_string[position]
        if letter == "-":
            symbol = unit_string[:position]
            number = unit_string[position + 1:]
            if "." not in number:
                exponent = -int(number)
            else:
                exponent = -float(number)
            break
        elif letter.isnumeric():
            if position == 0:
                if unit_string != '1':
                    raise UnitsError(f"Invalid unit string: {unit_string!r}")
                else:
                    return unit_string, 1
            symbol = unit_string[:position]
            number = unit_string[position:]
            if "." not in number:
                exponent = int(number)
            else:
                exponent = float(number)
            break
    return symbol, exponent


''' --------------------- UnitsStringManager --------------------- '''

class UnitsStringManager(dict):
    '''manages string of units; keys are unit symbol bases, values are exponents.'''
    bases = SI_BASE_UNIT_SYMBOLS
    known = KNOWN_CONVERSIONS

    def _get_bases_and_known(self, bases=None, known=None):
        '''return bases & known, using value from self if None is entered'''
        if bases is None: bases = self.bases
        if known is None: known = self.known
        return bases, known

    @classmethod
    def from_string(cls, string_of_units, *, bases=None, known=None):
        '''new UnitsStringManager from a string of units; just gets the symbols & exponents.'''
        units = string_of_units.split()
        symbols_and_exponents = [_get_symbol_and_exponent(unit) for unit in units]
        totals = dict()
        for s, exp in symbols_and_exponents:
            totals[s] = totals.get(s, 0) + exp
        result = cls(totals)
        if bases is not None: result.bases = bases
        if known is not None: result.known = known
        return result

    def _new(self, dict_):
        '''create new instance of self given this dict.'''
        return type(self)(dict_)

    def __pow__(self, exp):
        '''raise self to exponent; multiplies each value in self by exp.'''
        return self._new({key: value * exp for key, value in self.items()})

    def __mul__(self, other):
        '''multiply self by other UnitsStringManager; adds values where keys match.'''
        result = self.copy()
        for key, value in other.items():
            if key in result:
                result[key] = result[key] + value
            else:
                result[key] = value
        return result

    def __eq__(self, other):
        '''tells whether self equals another UnitsStringManager.'''

    def clean(self):
        '''return result of deleting all keys with value == 0.'''
        return self._new({key: value for key, value in self.items() if value != 0})

    def copy(self):
        '''return copy of self.'''
        return self._new(super().copy())

    def expand_once(self, *, bases=None, known=None):
        '''returns result of replacing all keys other than bases with their known conversions.'''
        bases, known = self._get_bases_and_known(bases, known)
        result = self._new({})  # empty dict.
        if len(self.items()) == 1 and DIMENSIONLESS in self:
            return result  # return empty dict; self only has DIMENSIONLESS in it.
        for key, value in self.items():
            if key in bases:
                result = result * self._new({key: value})
            elif key in known:
                result = result * (self.from_string(known[key]) ** value)
            else:
                raise UnitsError(f"Unknown unit: {key!r}")
        return result

    def expand(self, *, bases=None, known=None):
        '''returns result of expanding repeatedly until all keys are si base units.'''
        bases, known = self._get_bases_and_known(bases, known)
        result = self
        safety = tuple(self.items())
        while not all(key in bases for key in result.keys()):
            result = result.expand_once(bases=bases, known=known)
            if safety == tuple(result.items()):
                raise NotImplementedError('expand_once made no changes... raising error to stop endless loop.')
        return result

    def __str__(self):
        '''convert self into units string.'''
        self = self.clean()
        result = ' '.join([f'{key}{value}' for key, value in self.items()])
        return result

    def evaluate_partial(self, values_dict):
        '''evaluates all the parts of self which can be evaluated based on values_dict.
        returns (numerical value, UnitsStringManager with still-unevaluated units)

        if no keys in self are also in values_dict, result[0] will be 1.
        if all keys in self are also in values_dict, len(result[1]) will be 0.

        if values_dict provides any relevant key with value None, returns (None, {})
        '''
        value = 1
        unevaluated = self.copy()
        for key, exponent in self.items():
            if key in values_dict:
                vdval = values_dict[key]
                if vdval is None:
                    return None, type(self)()
                value = value * values_dict[key] ** exponent
                del unevaluated[key]
        return value, unevaluated

    def evaluate_iteratively(self, values_dict):
        '''evaluates all the parts of self which can be evaluated based on values_dict,
        and also calling result[1].expand_once() after each incomplete evaluation attempt.
        returns numerical value.

        raises UnitsError if any units are not in values_dict after expanding as many times as possible.
        '''
        result, units = self.evaluate_partial(values_dict)
        while len(units) > 0:
            u_expanded = units.expand_once()
            if u_expanded == units:
                raise UnitsError(f'Could not fully evaluate {self!r}; missing values for {list(units.keys())}.')
            next_number, units = u_expanded.evaluate_partial(values_dict)
            if next_number is None:
                return None
            result = result * next_number
        return result

    def evaluate(self, values_dict):
        '''evaluates self based on values_dict, returning a numerical value.
        if any units are not in values_dict, raises UnitsError.

        Equivalent to self.evaluate_partial(values_dict)[0], when all units are in values_dict.
        '''
        result = self.evaluate_partial(values_dict)
        if len(result[1]) > 0:
            raise UnitsError(f'Could not fully evaluate {self!r}; missing values for {list(result[1].keys())}.')
        return result[0]


def units_string_to_si_bases(string_of_units, *, bases=SI_BASE_UNIT_SYMBOLS, known=KNOWN_CONVERSIONS):
    '''convert string of any units to string of SI base units.'''
    return str(UnitsStringManager.from_string(string_of_units, bases=bases, known=known).expand())


''' --------------------- UnitsManager --------------------- '''

_units_manager_helpstr = \
    '''BASES tells all the SI base unit symbols.
    KNOWN tells all the known conversions between symbols (just strings, not numerical values).
    PHYSICAL_CONSTANTS_SI tells all the known physical constants, as tuples of (SI value, units string)).
    physical_constants_raw (or pcr) tells all the known physical constants, in 'raw' units;
        first do self.populate_pcr() to calculate them all.

    self.sic tells all the known values for unit conversions, from raw to SI.
        e.g., Mass [raw] * self.sic['M'] == Mass [si].
        new conversion factors are cached here the first time they are calculated,
            to avoid recalculating the values multiple times.
        to remove all cached values, self.clear_sic().

    call the object to get conversion factor. E.g.:
        density_si = density_raw * self('r', 'si', 'raw')
        density_raw = density_si * self('r', 'raw', 'si')
    or call the object to get physical constant. E.g.:
        c_si = self('c', 'si')   # speed of light in 'si' units
        c_raw = self('c', 'raw') # speed of light in 'raw' units.

    Conversion factors can be combined and raised to powers, e.g.:
        self('u0.5 m-1') == self('u')**0.5 * self('m')**-1
    '''

@format_docstring(helpstr=_units_manager_helpstr, sub_indent=DEFAULTS.TAB)
class UnitsManager():
    '''Manages units, including conversion factors and physical constants in different unit systems.

    :param units: str
        the default unit system to convert to ("si", "raw", or "cgs")
    :param M: value, default 1.        Mass_si = M * Mass_raw
    :param l: value, default 1.      Length_si = l * Length_raw
    :param t: value, default 1.        Time_si = t * Time_raw
    :param q: value, default 1.      Charge_si = q * Charge_raw
    :param K: value, default 1. Temperature_si = K * Temperature_raw
    :param sic_quantities: iterable
        calculate all these conversion factors & shorthands once, ahead of time.
        if None, use cls.SIC_QUANTITIES (default empty list)

    To alter cgs unit conversions (e.g. to specify q units), make a new self.CGS_UNITS.
        The default is CGS_UNITS = UnitsManager(M=1e-3, l=1e2, t=1, q=None, K=1).
        (don't alter the default one directly; make a new one with the desired values.)
        q=None is the default because electromagnetic cgs units are ambiguous,
            there are multiple options for how to convert to si and the equations differ.

    --- UnitsManager.help() will print a helpful message: ---
        {helpstr}
    '''
    BASES = SI_BASE_UNIT_SYMBOLS
    KNOWN = KNOWN_CONVERSIONS
    PHYSICAL_CONSTANTS_SI = PHYSICAL_CONSTANTS_SI
    pcr = alias('physical_constants_raw')
    SIC_QUANTITIES = []

    _helpstr = _units_manager_helpstr

    @classmethod
    def help(cls):
        '''prints a helpful message about using cls.'''
        print(f'Help for {cls.__name__}:\n{cls._helpstr}')

    # # # INITIALIZATION # # #
    def __init__(self, units="si", *, M=1, l=1, t=1, q=1, K=1, sic_quantities=None):
        self.units = units
        self.bases = dict(M=M, l=l, t=t, q=q, K=K)
        self._init_sic_quantities = sic_quantities
        self._init_from_bases()

    def _init_from_bases(self):
        '''initialize all other parts of self, using self.bases (and self._init_sic_quantities).'''
        self.init_sic()
        self.populate_sic(self._init_sic_quantities)
        self.populate_physical_constants_raw([])

    def init_sic(self):
        '''initialize self.sic from self.bases.'''
        b = self.bases
        A = None if ((b['q'] is None) or (b['t'] is None)) else b['q'] / b['t']
        self.sic = dict(kg=b['M'], m=b['l'], s=b['t'], A=A, K=b['K'])

    def clear_sic(self):
        '''clear self.sic. delete self.sic, then do self.init_sic()'''
        del self.sic
        self.init_sic()

    alts = simple_property('_alts', setdefault=dict,
        doc='''dict of {key: units} for alternative objects' unit systems.
        e.g. CoordsUnitsHaver might set alts['coords'] = 'si' to indicate coords always in si.''')

    # # # BASES # # #
    def _base_unit_property(ustr, longname):
        '''create property for base unit -- link to self.bases[ustr]. re-initialize self if changed.'''
        def getter(self):
            return self.bases[ustr]
        def setter(self, value):
            self.bases[ustr] = value
            self._init_from_bases()
        return property(getter, setter, doc=f'''{longname}_si = {ustr} * {longname}_raw''')

    M = _base_unit_property('M', 'Mass')
    l = _base_unit_property('l', 'Length')
    t = _base_unit_property('t', 'Time')
    q = _base_unit_property('q', 'Charge')
    K = _base_unit_property('K', 'Temperature')

    def is_trivial(self):
        '''returns whether self has only trivial conversion factors in it (all factors 1 or not provided).'''
        return all(val is None or val==1 for val in self.bases.values())

    # # # CGS UNITS # # #
    CGS_UNITS = simple_property('_CGS_UNITS', setdefaultvia='CGS_UNITS_DEFAULT',
            doc='''UnitsManager for cgs units. Default is UnitsManager(M=1e-3, l=1e2, t=1, q=None, K=1).
            q=None is the default because electromagnetic cgs units are ambiguous,
                there are multiple options for how to convert to si and the equations differ.
            Note that for CGS_UNITS, 'raw' means 'cgs'.
            Feel free to set self.CGS_UNITS = a new UnitsManager with a known q,
                if you have decided on a specific cgs system to remove relevant ambiguities.''')

    @classmethod
    def CGS_UNITS_DEFAULT(cls):
        '''UnitsManager for cgs units. Default is UnitsManager(M=1e-3, l=1e2, t=1, q=None, K=1).
        q=None is the default because electromagnetic cgs units are ambiguous,
            there are multiple options for how to convert to si and the equations differ.
        Note that for the CGS_UNITS result,'raw' means 'cgs'.
        '''
        # M=1e-3 because 1 g == 1e-3 kg. --> Mass [si] = 1e-3 * Mass [raw]
        # l=1e-2 because 1 cm == 1e-2 m. --> Length [si] = 1e-2 * Length [raw]
        return cls(M=1e-3, l=1e-2, t=1, q=None, K=1)

    # # # CONVERTING UNITS # # #
    def __call__(self, ustr, units=UNSET, convert_from='raw', *, alt=None):
        ''' return quantity or conversion factor in the specified unit system.
        ustr: str
            physical constant or quantity (e.g. "c" for speed of light or "M" for mass units)
        units: UNSET, None, 'si', 'raw', or 'cgs'
            unit system to convert to.
            UNSET --> units = self.alts.get(alt, self.units). (Equivalent to None if alt is None)
            None --> use self.units.
        convert_from: 'si', 'raw', or 'cgs'
            unit system to convert from.
            Ignored if ustr is a quantity (e.g. "c") instead of conversion factor (e.g. "mass")
        alt: object (probably None or str)
            if units is UNSET, use units = self.alts.get(alt, self.units)
            otherwise, alt must be None.
        :return: a converted physical constant or a conversion factor
        '''
        if units is UNSET:
            units = self.alts.get(alt, self.units)
        elif alt is not None:
            raise InputConflictError(f'cannot provide both units & alt! got alt={alt!r}, units={units!r}')
        if units is None:
            units = self.units

        # IF PHYSICAL CONSTANT, GET VALUE & UNIT STR.
        if ustr in self.PHYSICAL_CONSTANTS_SI:
            if units=='si':
                return self.PHYSICAL_CONSTANTS_SI[ustr][0]
            elif units=='raw':
                pcr = self.physical_constants_raw
                try:
                    return pcr[ustr]
                except KeyError:
                    pass  # handled below to avoid layered error messages if crash
                value, new_ustr = self.PHYSICAL_CONSTANTS_SI[ustr]
                result = value * self(new_ustr, 'raw', convert_from='si')
                pcr[ustr] = result
                return result
            elif units=='cgs':
                cgsunits = self.CGS_UNITS
                return cgsunits(ustr, 'raw')  # for CGS_UNITS, 'raw' means 'cgs'.
            else:
                raise NotImplementedError(f'unit system {units!r}, for physical constant {ustr!r}')
        
        # ELSE, NOT A PHYSICAL CONSTANT. RETURN CONVERSION FACTOR
        unit_systems = set((units, convert_from))
        if convert_from == units:
            return 1
        if "raw" in unit_systems:
            raw_to_si = self.get_sic_factor(ustr)
            if "si" in unit_systems:  # raw & si
                if (units == "si") and (convert_from == "raw"):
                    return raw_to_si
                elif (units == "raw") and (convert_from == "si"):
                    return 1 / raw_to_si
                assert False, 'coding error if reached this line'
            elif "cgs" in unit_systems:  # raw & cgs
                cgsunits = self.CGS_UNITS
                si_to_cgs = cgsunits(ustr, 'raw', 'si')  # for CGS_UNITS, 'raw' means 'cgs'.
                raw_to_cgs = raw_to_si * si_to_cgs
                if (units == "cgs") and (convert_from == "raw"):
                    return raw_to_cgs
                elif (units == "raw") and (convert_from == "cgs"):
                    return 1 / raw_to_cgs
                assert False, 'coding error if reached this line'
        elif unit_systems == set(("si", "cgs")):
            cgsunits = self.CGS_UNITS
            if units == "si":   # and convert_from == "cgs"
                return cgsunits(ustr, 'si', 'raw')  # for CGS_UNITS, 'raw' means 'cgs'.
            else: # units == "cgs" and convert_from == "si"
                return cgsunits(ustr, 'raw', 'si')  # for CGS_UNITS 'raw' means 'cgs'.
        raise NotImplementedError(f'units conversion from {convert_from!r} to {units!r}')

    def calc_sic_factor(self, ustr):
        '''Calculate conversion factor from raw units to SI units.
        Also save result in self.sic for future use.
        :param ustr: string of units (e.g. "m-3")
        :return: conversion factor from raw units to SI units
        '''
        uobj = UnitsStringManager.from_string(ustr)
        result = uobj.evaluate_iteratively(self.sic)
        self.sic[ustr] = result
        return result

    def get_sic_factor(self, ustr):
        '''Get conversion factor from raw units to SI units.
        If not already calculated, calculate it first.
        :param ustr: string of units (e.g. "m-3")
        :return: conversion factor from raw units to SI units

        if None, raise UnitsUnknownError instead.
        '''
        sic = self.sic
        if ustr in sic:
            result = sic[ustr]
        else:
            result = self.calc_sic_factor(ustr)
        if result is None:
            errmsg = (f'Cannot determine units for {ustr!r}. This most likely means '
                      f'it depends on one or more base units which are unknown (i.e. None). '
                      f'Bases defined here are: {self.bases!r}.')
            raise UnitsUnknownError(errmsg)
        return result

    def string_to_si_bases(self, string_of_units):
        '''convert string of any units to string of SI base units.'''
        return units_string_to_si_bases(string_of_units, bases=self.BASES, known=self.KNOWN)

    # # # CALCULATING LOTS OF QUANTITIES # # #
    def populate_sic(self, quantities=None, *, reset=False):
        '''Calculate lots of conversion factors.
        :param quantities: None or iterable of strings
            calculate these conversion factors and put them in self.sic.
            e.g. "p", "momentum", "M u", "kg m s-1".
            if None, use self.SIC_QUANTITIES.
        :reset: bool, default False.
            if True, clear self.sic before calculating.
        :return: self.sic, after adding all the requested conversion factors.
        '''
        if quantities is None: quantities = self.SIC_QUANTITIES
        if reset: self.clear_sic()
        for quantity in quantities:
            self.get_sic_factor(quantity)
        return self.sic

    def populate_physical_constants_raw(self, quantities=None):
        '''Calculate all the physical constants in raw units.
        :param quantities: None or iterable of strings
            calculate only these physical constants.
            if None, use self.PHYSICAL_CONSTANTS_SI.keys()
        :return: self.physical_constants_raw, after creating and filling it.
        '''
        if quantities is None: quantities = self.PHYSICAL_CONSTANTS_SI.keys()
        pcr = {key: self(key, 'raw') for key in quantities}
        self._physical_constants_raw = pcr
        return self._physical_constants_raw

    @property
    def physical_constants_raw(self):
        '''dict of all the physical constants in raw units.'''
        try:
            return self._physical_constants_raw
        except AttributeError:
            return self.populate_physical_constants_raw()

    # # # DISPLAY # # #
    display_precision = 2  # precision for displaying numbers in repr(self)

    def _repr_show_factors(self):
        '''returns dict of name and conversion factor to si, to include in repr(self).
        Here, include l, t, M, and q. Also include K if it is not 1.
        '''
        factors = {key: self(key, 'si', 'raw') for key in ('l', 't', 'M', 'q', 'K')}
        if factors['K'] == 1:
            del factors['K']
        return factors

    def __repr__(self):
        prec = self.display_precision
        factors = self._repr_show_factors()
        factors = [f"{key}={value:.{prec}e}" for key, value in factors.items()]
        return f'{type(self).__name__}(units={self.units!r}, {", ".join(factors)})'



"""
File Purpose: test xarray save & xarray load.
"""
import os
import pytest

import numpy as np
import xarray as xr

import PlasmaCalcs as pc
pc.DEFAULTS.pic_ambiguous_unit = dict(u_t=1)  # must be defined before loading an EppicCalculator.

HERE = os.path.dirname(__file__)
TESTS_DIR = os.path.dirname(HERE)
ARTIFACTS = os.path.join(TESTS_DIR, '_test_artifacts', 'xarray_io')
if os.path.exists(ARTIFACTS):  # remove any xarray_io artifacts before running tests.
    import shutil; shutil.rmtree(ARTIFACTS)
# <-- don't makedirs here; xarray_save should makedirs as needed.

''' --------------------- can create eppic calculator --------------------- '''

def get_eppic_calculator(**kw_init):
    '''make an eppic calculator'''
    with pc.InDir(os.path.join(HERE, 'test_eppic_tinybox')):
        ec = pc.EppicCalculator.from_here(**kw_init)
    return ec


''' --------------------- load / save / load some basic stuff! --------------------- '''

@pytest.mark.filterwarnings('ignore:`product`.*NumPy.*:DeprecationWarning')
def test_eppic_xarray_io():
    '''test saving & loading xarray with DimensionValue coords.'''
    ec = get_eppic_calculator()

    # array without ANY DimensionValue coordinates.
    arr = ec('7')  # just the number 7, as a DataArray!
    dst = arr.pc.save(os.path.join(ARTIFACTS, 'just_7'))
    arr_loaded = pc.xarray_load(dst)
    assert np.all(arr == arr_loaded)

    # very small array
    n0 = ec('n0')
    dst = n0.pc.save(os.path.join(ARTIFACTS, 'n0'))
    n0_loaded = pc.xarray_load(dst)
    assert np.all(n0 == n0_loaded)

    # array with spatial extent
    n = ec('n')
    dst = n.pc.save(os.path.join(ARTIFACTS, 'n'))
    n_loaded = pc.xarray_load(dst)
    assert np.all(n == n_loaded)
    # check naming scheme:
    assert os.path.basename(dst) == 'n.pcxarr'
    assert dst == os.path.abspath(os.path.join(ARTIFACTS, 'n.pcxarr'))

    # crash if would overwrite but not exist_ok:
    with pytest.raises(FileExistsError):
        n.pc.save(os.path.join(ARTIFACTS, 'n'))
    # don't crash if exist_ok:
    n.pc.save(os.path.join(ARTIFACTS, 'n'), exist_ok=True)
    n_loaded = pc.xarray_load(dst)
    assert np.all(n == n_loaded)

    # array at only 1 snap.
    n = ec('n', snap=7)
    dst = n.pc.save(os.path.join(ARTIFACTS, 'n_at_snap=7'))
    n_loaded = pc.xarray_load(dst)
    assert np.all(n == n_loaded)

    # array with only 1 fluid. Also with multiple components
    u = ec('u', fluid=0)
    dst = u.pc.save(os.path.join(ARTIFACTS, 'u_at_fluid=0'))
    u_loaded = pc.xarray_load(dst)
    assert np.all(u == u_loaded)

    # array with only 1 fluid and only 1 component.
    u = ec('u', fluid=1, component='z')
    dst = u.pc.save(os.path.join(ARTIFACTS, 'u_at_fluid=1_component=z'))
    u_loaded = pc.xarray_load(dst)
    assert np.all(u == u_loaded)

    # dataset
    n_at_box_edges = ec('n', multi_slices=dict(ndim=1))
    assert isinstance(n_at_box_edges, xr.Dataset)
    dst = n_at_box_edges.pc.save(os.path.join(ARTIFACTS, 'n_at_box_edges'))
    n_at_box_edges_loaded = pc.xarray_load(dst)
    assert n_at_box_edges.equals(n_at_box_edges_loaded)

    # array with bool-valued attrs
    arr = ec('7')  # just the number 7, as a DataArray!
    arr.attrs['attr0'] = True
    arr.attrs['attr1'] = False
    dst = arr.pc.save(os.path.join(ARTIFACTS, 'array_with_bool_attrs'))
    arr_loaded = pc.xarray_load(dst)
    assert np.all(arr == arr_loaded)

    # dataset with bool-valued attrs
    ds = ec(['0', '7'])
    ds.attrs['attr0'] = True
    ds.attrs['attr1'] = False
    dst = ds.pc.save(os.path.join(ARTIFACTS, 'dataset_with_bool_attrs'))
    ds_loaded = pc.xarray_load(dst)
    assert ds.equals(ds_loaded)

    # dataset with data_vars that have bool-valued attrs
    ds = ec(['0', '7'])
    ds['0'].attrs['attr0'] = True
    ds['7'].attrs['attr1'] = False
    dst = ds.pc.save(os.path.join(ARTIFACTS, 'dataset_with_bool_attrs_in_data_vars'))
    ds_loaded = pc.xarray_load(dst)
    assert ds.equals(ds_loaded)

    # toplevel_scale_coords (and, saving array with simple dict-valued attr)
    arr = ec('moment1', toplevel_scale_coords={'t': 1000}, assign_behavior_attrs=True)
    assert np.isclose(arr['t'].isel(snap=3), ec.snaps[3].t * ec.u('t') * 1000)
    assert arr.attrs['toplevel_scale_coords'] == {'t': 1000}
    dst = arr.pc.save(os.path.join(ARTIFACTS, 'moment1_with_toplevel_scale_coords_t_1000'))
    arr_loaded = pc.xarray_load(dst)
    assert np.all(arr == arr_loaded)
    assert arr_loaded.attrs['toplevel_scale_coords'] == {'t': 1000}

    # saving & loading UniqueDimensionValue (here: snap is INPUT_SNAP)
    ec.set_vars_from_inputs()
    assert ec.snap is pc.INPUT_SNAP
    arr = ec('n')
    dst = arr.pc.save(os.path.join(ARTIFACTS, 'n_at_INPUT_SNAP'))
    arr_loaded = pc.xarray_load(dst)
    assert np.all(arr == arr_loaded)
    assert arr_loaded['snap'].item() is pc.INPUT_SNAP



"""
File Purpose: shorthand for quantities "perpendicular to B"
"""

from .quantity_loader import QuantityLoader

class BperpLoader(QuantityLoader):
    '''shorthand for quantities "perpendicular to B".'''
    @known_var(deps=['u_perp_B'])
    def get_uperp(self):
        '''velocity vector, perpendicular to B. This is a full 3-vector.
        Equivalent: self('u_perp_B') == u - self('u_par_B') == u - (u dot Bhat) Bhat
        '''
        return self('u_perp_B')

    @known_var(deps=['u_par_B'])
    def get_upar(self):
        '''velocity vector, parallel to B. This is a full 3-vector.
        Equivalent: self('u_par_B') == (u dot Bhat) Bhat
        '''
        return self('u_par_B')

    @known_var(deps=['E_perp_B'])
    def get_Eperp(self):
        '''electric field, perpendicular to B. This is a full 3-vector.
        Equivalent: self('E_perp_B') == E - self('E_par_B') == E - (E dot Bhat) Bhat
        '''
        return self('E_perp_B')

    @known_var(deps=['E_par_B'])
    def get_Epar(self):
        '''electric field, parallel to B. This is a full 3-vector.
        Equivalent: self('E_par_B') == (E dot Bhat) Bhat
        '''
        return self('E_par_B')

    @known_pattern(r'Ta(joule)?perp', deps=[{0: 'Ta{group0}_perp_B'}])
    def get_Taperp(self, var, *, _match=None):
        '''Taperp --> anisotropic temperature, perpendicular to B. This is a full 3-vector.
        Equivalent: self('Ta_perp_B') == Ta - self('Ta_par_B') == Ta - (Ta dot Bhat) Bhat
        Also supports 'Tajouleperp' to get value in energy units; == 'Taperp*kB'.
        '''
        joule, = _match.groups()
        if joule is None: joule = ''
        return self(f'Ta{joule}_perp_B')

    @known_pattern(r'T(joule)?par', deps=[{0: 'Ta{group0}_par_B'}])
    def get_Tapar(self, var, *, _match=None):
        '''Tapar --> anisotropic temperature, parallel to B. This is a full 3-vector.
        Equivalent: self('Ta_par_B') == (Ta dot Bhat) Bhat
        Also supports 'Tajoulepar' to get value in energy units; == 'Tapar*kB'.
        '''
        joule, = _match.groups()
        if joule is None: joule = ''
        return self(f'Ta{joule}_par_B')



"""
File Purpose: plotting routines by Save
"""

import matplotlib.pyplot as plt

from ....plotting import PlotterManager


class EppicPlotterManagerSave(PlotterManager):
    '''eppic plotting routines by Save.
    Consider using EppicCalculator.make_plots(who='save') to make all the plots.
    '''

    # remember to say who='save' for each of these!
    @known_plotter(who='save')
    def plot_moment1_specie0(ec, **kw_plot_settings):
        '''plot the first moment of the simulation for specie0'''
        result = ec('moment1', fluid=0).pc.timelines(**kw_plot_settings)
        plt.title(f'{ec.title}, species={ec.fluids.get(0)}')
        return result



"""
File Purpose: The main calculator class, for hookups to inherit from.

Subclassing order recommendation:
    put the PlasmaCalculator class LAST in the order, to ensure methods from subclass are used.
    E.g. class HookupCalculator(FooLoader, BarLoader, QuuxHaver, PlasmaCalculator).
"""

from .addons import AddonLoader
from .dimensions import (
    ComponentHaver, FluidsHaver, SnapHaver,
    MainDimensionsHaver,
)
from .plotting import PlotterManager
from .quantities import (
    # bases:
    AllBasesLoader, SimpleDerivedLoader,
    # misc:
    BperpLoader,
    CollisionsLoader,
    ConstantsLoader,
    MaskLoader,
    PlasmaDriftsLoader, PlasmaHeatingLoader, PlasmaParametersLoader,
    PlasmaStatsLoader,
    QuasineutralLoader,
    TimescalesLoader,
    # patterns, especially stats, arithmetic, calculus:
    ParenthesisLoader,
    CachesLoader,
    BlurLoader,
    FFTLoader,
    FluidsLoader,
    PolyfitLoader,
    StatsLoader,
    BasicArithmeticLoader, BasicDerivativeLoader,
    ComparisonLoader,
    VectorArithmeticLoader, VectorDerivativeLoader,
)
from .units import UnitsHaver, CoordsUnitsHaver

class DimensionlessPlasmaCalculator(
                        UnitsHaver,   # units come first since they are used by some loaders
                        ParenthesisLoader,   # check for parenthesis before other patterns
                        BperpLoader,
                        ConstantsLoader,
                        # drifts & heating inherits from parameters loader, so they must come first:
                        PlasmaDriftsLoader, PlasmaHeatingLoader, PlasmaParametersLoader,
                        PlasmaStatsLoader,  # inherits from StatsLoader, so must come first.
                        QuasineutralLoader,
                        TimescalesLoader,
                        # bases loaders go last, in case other vars loaders override:
                        AllBasesLoader, SimpleDerivedLoader,
                        # pattern-based loaders go after bases loaders.
                        CachesLoader,
                        BlurLoader,
                        FFTLoader,
                        PolyfitLoader,
                        StatsLoader,
                        BasicArithmeticLoader, BasicDerivativeLoader,
                        ComparisonLoader,
                        MaskLoader,
                        # addons go last; addons can provide extra methods but not override existing
                        AddonLoader,
                        # plotter manager can go anywhere
                        PlotterManager,
                       ):
    '''class for plasma calculator object.

    Not intended for direct instantiation. Instead, see options in the "hookups" subpackage,
        or write your own hookup for a different type of input, following the examples there.
    '''
    # note: the order of Loader classes *does* matter;
    #  will look in earlier classes for implementations first.
    #  e.g. if PlasmaCalculator(..., classA, ..., classB, ...) and classA and classB
    #       both have a get_foo method, then classA.get_foo will be used.
    pass

class ComponentHavingPlasmaCalculator(
            ComponentHaver,
            DimensionlessPlasmaCalculator,
            # vector loaders intentionally after DimensionlessPlasmaCalculator,
            #   so that they will be after basic arithmetic.
            #   E.g., this way log10_E_par_B gets log10_(E_par_B), not (log10_E)_par_B.
            VectorArithmeticLoader, VectorDerivativeLoader,
            ):
    '''DimensionlessPlasmaCalculator but with vector arithmetic and derivatives.

    Not intended for direct instantiation. Instead, see options in the "hookups" subpackage,
        or write your own hookup for a different type of input, following the examples there.
    '''
    pass

class PlasmaCalculator(CoordsUnitsHaver, MainDimensionsHaver, SnapHaver,
                       ComponentHavingPlasmaCalculator):
    '''DimensionlessPlasmaCalculator but added dimensions: component, main dimensions, snaps.

    Not intended for direct instantiation. Instead, see options in the "hookups" subpackage,
        or write your own hookup for a different type of input, following the examples there.
    '''
    pass

class MultifluidPlasmaCalculator(CollisionsLoader, FluidsLoader, FluidsHaver,
                                 PlasmaCalculator):
    '''PlasmaCalculator, also with fluid and jfluid.

    Not intended for direct instantiation. Instead, see options in the "hookups" subpackage,
        or write your own hookup for a different type of input, following the examples there.
    '''
    # CollisionsLoader goes here since it depends on jfluid.
    pass


''' --------------------- Niche combinations, sometimes useful --------------------- '''

class VectorlessPlasmaCalculator(CoordsUnitsHaver, MainDimensionsHaver, SnapHaver,
                                 DimensionlessPlasmaCalculator):
    '''PlasmaCalculator without vector components/arithmetic/derivatives.
    Has all other features of PlasmaCalculator (e.g. main dimensions, snaps)

    Not intended for direct instantiation. Instead, see options in the "hookups" subpackage,
        or write your own hookup for a different type of input, following the examples there.
    '''
    pass

class VectorlessMultifluidPlasmaCalculator(CollisionsLoader, FluidsLoader, FluidsHaver,
                                           VectorlessPlasmaCalculator):
    '''VectorlessPlasmaCalculator, also with fluid and jfluid.
    same as MultifluidPlasmaCalculator except without vector components / arithmetic.

    Not intended for direct instantiation. Instead, see options in the "hookups" subpackage,
        or write your own hookup for a different type of input, following the examples there.
    '''
    pass



"""
File Purpose: multifluid analysis from muram
"""
import os

from .muram_calculator import MuramCalculator
from .muram_multifluid_densities import MuramMultifluidDensityLoader
from ...dimensions import SINGLE_FLUID
from ...mhd import MhdMultifluidCalculator, SpecieList
from ...tools import format_docstring

@format_docstring(muram_calculator_docs=MuramCalculator.__doc__, sub_ntab=1)
class MuramMultifluidCalculator(MuramMultifluidDensityLoader,
                                MhdMultifluidCalculator,
                                MuramCalculator):
    '''MultifluidPlasmaCalculator for MURAM outputs.
    various possible ways to infer fluids.
        One possibility is to use abundances + saha ionization equation.
        [TODO] explain this in more detail.
        [TODO] allow to enter fluids as list of strings during init.

    set self.fluid=SINGLE_FLUID to get single-fluid values,
        otherwise will get inferred multifluid values.

    --- Docstring from MuramCalculator copied below ---
        {muram_calculator_docs}
    '''

    SINGLE_FLUID = SINGLE_FLUID  # convenient reference to SINGLE_FLUID. Subclass should NEVER override.

    def __init__(self, *, dir=os.curdir, units='si', **kw_super):
        super().__init__(dir=dir, units=units, **kw_super)
        self.init_fluids()

    # # # FLUIDS # # #
    def init_fluids(self):
        '''initialize self.fluids, fluid, jfluids, and jfluid.'''
        self.fluids = self.chromo_fluid_list()
        self.fluid = None
        self.jfluids = self.fluids
        self.jfluid = self.jfluids.get_neutral()

    def chromo_fluid_list(self):
        '''SpecieList of species relevant to the chromosphere, maybe.
        currently, just produces list of [electron, H_I, H_II, *other_once_ionized_ions]
        '''
        elements = self.tabin.elements
        if (elements[0] != 'H'):
            raise NotImplementedError('[TODO] chromo_fluid_list for other elements list?')
        H = elements.get('H')
        other = elements[1:]
        result = [SpecieList.value_type.electron(),
                  H.neutral(),
                  H.ion(1),
                  ]
        result.extend(other.ion_list(q=1))
        return SpecieList(result, istart=0)  # istart=0 --> renumber for this list.



"""
File Purpose: caching quantities

this is non-trivial because they need to be associated with behavior attrs.
E.g., cached value for 'n' may be associated with 'fluid', 'snap', 'units', ...

additionally, note that set_var may cache a different var than the one entered,
if the entered var has a known_setter.
E.g. for eppic, set_var('n', fluid=1) may cache a value for 'den1' instead of 'n'.

[TODO] detect when behavior is a relevant sublist of dimensions of another behavior,
    e.g. quantity with fluid=1 should be incorporated into getting value with fluid=[0,1,2,3].
    and, quantity with fluid=[0,1,2,3] should be incorporated into getting value with fluid=1.
    The main code can then check relevant_sublist, and if one is available,
        load_across_dims() for the dims with relevant sublists in cache.
"""
import collections

from ..dimensions import BehaviorQuantity
from ..errors import CacheNotApplicableError
from ..tools import (
    simple_property, alias,
    format_docstring,
)
from ..defaults import DEFAULTS


@format_docstring(behavior_doc=BehaviorQuantity.__doc__.replace('BehaviorQuantity', 'CachedQuantity'))
class CachedQuantity(BehaviorQuantity):
    '''{behavior_doc}
    time: number, default 0
        amount of time it took to load this quantity originally.
        Not required for any caching behavior, but helpful for knowing if cache is useful.
    '''
    def __init__(self, value, behavior, *, MB=None, ukey=None, time=0, **kw__behavior_quantity):
        super().__init__(value, behavior, ukey=ukey, **kw__behavior_quantity)
        self.time = time
        self.n_used = 0  # counts number of times this value was loaded from cache.
                         # not updated by this class, but see e.g. VarCache.

    def _new_with_dims(self, dims, **kw_init):
        '''return new CachedQuantity like self but with dims instead of self.behavior.dims.'''
        kw_init.setdefault('time', self.time)
        return super()._new_with_dims(dims, **kw_init)

    def _repr_contents(self):
        contents = super()._repr_contents()
        if self.time is not None:
            contents.append(f'time={self.time:.2e}')
        return contents


class VarCache():
    '''cache of vars. {var: {id: CachedQuantity}.
    id is quant.id for CachedQuantity quant;
        it helps when determining which cached quantity to pop, if cache size is limited.

    Use self.append_cq(var, CachedQuantity) to add a CachedQuantity to this cache.
    Use self.append(var, value, behavior) to create a CachedQuantity and add it to this cache.

    Tracks self.order_entered (updated only when entering a new var),
        and self.order_used (updated when entering OR loading a var).
        Each of these is an OrderedDict with keys (var, qid) tuples; values None.
        The 'latest' entry is at the end.

    MBmax: None or number
        maximum size in MB of this cache. None --> no maximum.
        if provided along with nmax, the more restrictive one will be used.
    nmax: None or number
        maximum number of entries in this cache. None --> no maximum.
        if provided along with MBmax, the more restrictive one will be used.
    remove_strategy: 'used' or 'entered'
        whether to remove the oldest entries based on order_used or order_entered.
        Only applies when cache size exceeds max (determined by MBmax and nmax).
    '''
    def __init__(self, *, MBmax=None, nmax=None, remove_strategy='used'):
        self.cache = dict()
        self.MBmax = MBmax
        self.nmax = nmax
        self.remove_strategy = remove_strategy
        self.order_entered = collections.OrderedDict()
        self.order_used = collections.OrderedDict()
        super().__init__()

    # # # DICT-LIKE BEHAVIOR IN CACHE # # #
    def keys(self):
        '''return keys of self.cache.'''
        return self.cache.keys()

    def values(self):
        '''return values of self.cache.'''
        return self.cache.values()

    def items(self):
        '''return items of self.cache.'''
        return self.cache.items()

    def __getitem__(self, key):
        '''return self.cache[key].'''
        return self.cache[key]

    def __iter__(self):
        '''iterate over self.cache.'''
        return iter(self.cache)

    def __contains__(self, var):
        '''return whether at least one cached value for var exists in cache.'''
        return var in self.cache

    def __len__(self):
        '''return len of cache. This is the number of vars each with at least 1 cached value.'''
        return len(self.cache)

    def clear(self):
        '''empties the cache of all cached values.'''
        while len(self) > 0:
            self.pop()

    # # # ITERATION # # #
    def iter_quantities(self):
        '''iterate over all quantities in this cache.'''
        for qdict in self.values():
            for quant in qdict.values():
                yield quant

    # # # ADDING / REMOVING QUANTITIES # # #
    def append_cq(self, var, cached_quantity):
        '''add cached_quantity to this cache.
        if self.nmax or self.MBmax is provided, remove entries until size is below max.
        if cached_quantity is compatible with any existing cached quantity for var,
            delete the pre-existing one(s) before adding the new one.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        self._remove_entries_if_too_large(cached_quantity)
        varcache = self.cache.setdefault(var, {})
        if len(varcache) > 0:  # first, remove any compatible values already in varcache.
            for qid, quant in tuple(varcache.items()):
                if quant.matches_behavior(cached_quantity.behavior):
                    self.pop((var, qid), _clear_empty_var=False)
        varcache[cached_quantity.id] = cached_quantity
        self.order_entered[(var, cached_quantity.id)] = None
        self.order_used[(var, cached_quantity.id)] = None
        return cached_quantity

    def append(self, var, value, behavior, ukey=None, **kw):
        '''create CachedQuantity(value, behavior, ukey=ukey, **kw), and append it to this cache.
        var: str
        value: any
        behavior: dict or Behavior
            (if dict, internally convert to Behavior.)
        ukey: None or str
            if provided, pop 'units' from behavior, and ukey is the string to use
                for getting the conversion factor for value to another units system.
            See help(self) for details.
        '''
        return self.append_cq(var, CachedQuantity(value, behavior, ukey=ukey, **kw))

    @property
    def size(self):
        '''number of entries in this cache.'''
        return sum(len(d) for d in self.values())

    @property
    def size_MB(self):
        '''total size of this cache, in MB.'''
        # [TODO][EFF] keep a running total instead of recalculating.
        #   (probably not worthwhile to implement though; probably won't save much time.)
        return sum(quant.MB for quant in self.iter_quantities())

    MB = alias('size_MB')

    def _remove_entries_if_too_large(self, new_cached_quantity=None):
        '''remove entries from this cache until size WILL be below max AFTER adding new_cached_quantity.
        This method doesn't add new_cached_quantity, but pretends it will be added.
        if new_cached_quantity is None, clean up this cache without adding anything.

        returns (total number of entries removed due to nmax, number removed due to MBmax).
            Note - always attempts removals due to nmax first.
        '''
        nmax_remove = self._remove_entries_if_too_large_nmax(new_cached_quantity)
        MBmax_remove = self._remove_entries_if_too_large_MBmax(new_cached_quantity)
        return nmax_remove, MBmax_remove

    def _remove_entries_if_too_large_nmax(self, new_cached_quantity=None):
        '''remove entries from this cache until size is below nmax.
        if self.nmax is not provided, do nothing.
        if new_cached_quantity is provided, treat self.size as if it is 1 bigger.
        returns total number of entries removed.
        '''
        nmax = self.nmax
        if nmax is None:
            return
        account_for_new = 0 if new_cached_quantity is None else 1
        counter = 0
        while (self.size + account_for_new) > nmax:
            self.pop()
            counter += 1
        return counter

    def _remove_entries_if_too_large_MBmax(self, new_cached_quantity=None):
        '''remove entries from this cache until size is below MBmax.
        if self.MBmax is not provided, do nothing.
        if new_cached_quantity is provided, treat self.size_MB as if it is new_cached_quantity.MB bigger.
        returns total number of entries removed.
        '''
        MBmax = self.MBmax
        if MBmax is None:
            return
        account_for_new = 0 if new_cached_quantity is None else new_cached_quantity.MB
        counter = 0
        while (self.size_MB + account_for_new) > MBmax:
            self.pop()
            counter += 1
        return counter

    def pop(self, itemkey=None, *, strategy=None, _clear_empty_var=True):
        '''remove and return an entry from this cache.
        When called with no arguments, e.g. self.pop(), removes the oldest entry.

        itemkey: None or (var, cached_quantity.id) tuple
            if provided, pop THIS entry from the cache. (self.cache[var][cached_quantity.id])
            else, pop oldest entry.
        strategy: None, 'used', or 'entered'
            determines meaning of 'oldest'. If None, use self.remove_strategy.
        _clear_empty_var: bool
            whether to delete self.cache[var] if it becomes empty due to popping this item.
        '''
        if itemkey is None:
            itemkey = self._get_oldest_itemkey(strategy=strategy)
        # actually remove the item
        del self.order_used[itemkey]
        del self.order_entered[itemkey]
        var, qid = itemkey
        cached_quantity = self.cache[var].pop(qid)
        if _clear_empty_var and len(self.cache[var]) == 0:
            del self.cache[var]
        return cached_quantity

    def pop_used(self):
        '''remove and return the oldest entry from this cache, based on self.order_used.'''
        self.pop(strategy='used')

    def pop_entered(self):
        '''remove and return the oldest entry from this cache, based on self.order_entered.'''
        self.pop(strategy='entered')

    def _get_oldest_itemkey(self, strategy=None):
        '''return (var, cached_quantity.id) for the oldest item in this cache.
        strategy: None, 'used', or 'entered'
            determines meaning of 'oldest'. If None, use self.remove_strategy.
        '''
        if strategy is None: strategy = self.remove_strategy
        if strategy == 'used':
            ordering = self.order_used
        elif strategy == 'entered':
            ordering = self.order_entered
        else:
            raise ValueError(f'invalid remove_strategy {strategy!r}.')
        return next(iter(ordering))

    def mark_used(self, var, quant):
        '''mark (var, quant) as having just been used.
        This affects bookkeeping, as well as pop order for strategy='used', if cache gets too large.
        '''
        self.order_used.move_to_end((var, quant.id))
        quant.n_used += 1
        self.n_used += 1
        self.time_saved += quant.time

    def remove(self, var, behavior, *, missing_ok=False, **kw):
        '''remove all quantities for which behavior.compatible(quantity.behavior, lenient=True), for var.
        i.e., quantities where behavior[k]==quantity[k] for all k in [*behavior.keys(), *behavior.dims.keys()].

        missing_ok: bool
            if not missing_ok, raise CacheNotApplicableError if there are no relevant quantities.
        additional kwargs are passed to behavior.compatible(..., **kw). E.g. could use lenient=False instead.
        returns list of all removed quantities.
        '''
        kw.setdefault('lenient', True)
        kw.setdefault('subdims', True)
        quants_to_remove = self.inverse_lookup(var, behavior, **kw)
        if (not missing_ok) and (len(quants_to_remove) == 0):
            errmsg = f'no cached quantity with behavior.compatible(quantity.behavior), for var {var!r}.'
            raise CacheNotApplicableError(errmsg)
        result = []
        for quant in quants_to_remove:
            popped = self.pop((var, quant.id))
            result.append(popped)
        return result

    # # # TIMING STATS (to learn how useful the cache is being) # # #
    time_saved = simple_property('_time_saved', setdefault=lambda: 0,
                    doc='''total time saved by using this cache, in seconds.''')

    n_used = simple_property('_n_used', setdefault=lambda: 0,
                    doc='''total number of times a value was loaded from this cache.''')

    @property
    def time_saved_here(self):
        '''total time saved by quantities which are currently in the cache.'''
        return sum(quant.time * quant.n_used for quant in self.iter_quantities())

    @property
    def n_used_here(self):
        '''total number of times loaded quantities currently in the cache.'''
        return sum(quant.n_used for quant in self.iter_quantities())

    def timing(self):
        '''return dict of timing stats.'''
        return dict(time_saved=self.time_saved, n_used=self.n_used,
                    time_saved_here=self.time_saved_here, n_used_here=self.n_used_here)

    # # # DISPLAY # # #
    def __repr__(self):
        contents = []
        contents.append(f'size={self.size}')
        if self.nmax is not None:
            contents.append(f'nmax={self.nmax}')
        contents.append(f'MB={self.MB:.2e}')
        if self.MBmax is not None:
            contents.append(f'MBmax={self.MBmax}')
        contents.append(f'keys={list(self.keys())}')
        contents_str = ', '.join(contents)
        return f"{type(self).__name__}({contents_str})"

    # # # RETRIEVING QUANTITIES # # #
    def lookup(self, var, behavior):
        '''returns (relevance, quant) for cached quantity for var, with quantity.relevant(behavior).
        returns (False, None) if no quantity.relevant(behavior).
        else, returns (True, quant) if exact match, or (subdims, quant) if subdim match.
        
        subdim match occurs when dim from cached quantity is a value within dim from behavior
            e.g. fluid=0 in cached quantity, fluid=[0,1,2] in behavior.
        '''
        try:
            qdict = self.cache[var]
        except KeyError:
            return (False, None)
        for quant in qdict.values():
            relevance = quant.relevant(behavior)
            if relevance:
                return (relevance, quant)
        return (False, None)

    def inverse_lookup(self, var, behavior, *, lenient=True, subdims=True, **kw_compatible):
        '''returns list of quantities from self.cache[var] for which behavior.compatible(quantity.behavior)'''
        result = []
        try:
            qdict = self.cache[var]
        except KeyError:
            return result  # cache[var] does not exist.
        for quant in qdict.values():
            if quant.compatible(behavior, lenient=lenient, subdims=subdims, **kw_compatible):
                result.append(quant)
        return result

    def get(self, var, behavior):
        '''return cached value from CachedQuantity with matching behavior, if possible,
        else raise CacheNotApplicableError.

        behavior is matching if CachedQuantity.matches_behavior(behavior),
            i.e. if CachedQuantity.behavior.compatible(behavior, lenient=True).
        '''
        try:
            qdict = self.cache[var]
        except KeyError:
            raise CacheNotApplicableError(f'var {var!r} not in cache.')
        for quant in qdict.values():
            if quant.matches_behavior(behavior):
                # found relevant quant! do some bookkeeping, then return it.
                self.mark_used(var, quant)
                return quant.value
        errmsg = f'no quantity matching behavior, for var {var!r} in cache.'
        raise CacheNotApplicableError(errmsg)


@format_docstring(varcache_doc=VarCache.__doc__)
class VarCacheSingles(VarCache):
    '''cache of vars, where each cached quantity represents a single point (in dimension space).
    "single point" refers to having at most one value for each of the dimensions,
        but not maindims. E.g., one snap, fluid, component. Any number of x,y,z.

    When appending a CachedQuantity which is not a single point (in dimension space),
        separate it into multiple quantities, and append each of those instead.

    VarCache documentation:
    -----------------------
    {varcache_doc}
    '''
    def append_cq(self, var, cached_quantity):
        '''add cached_quantity to this cache.
        If cached_quantity is not a single point (in dimension space), first separate, then append each.

        if self.nmax or self.MBmax is provided, remove entries until size is below max.
        if cached_quantity is compatible with any existing cached quantity for var,
            delete the pre-existing one(s) before adding the new one.

        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        quant_at_points = cached_quantity.list_points()
        for quant in quant_at_points:
            super().append_cq(var, quant)
        return quant_at_points


"""
File Purpose: make it easier to deal with renaming fft dimensions

Easiest would be if always 'dim' --> 'freq_{dim}'.
However, dimname should include 'rad' if rad=True.
Additionally, there are well-known custom names for some rad fft dimensions,
    e.g. when rad=True, 'x' --> 'k_x', and 't' --> 'omega'.
"""

from ...errors import DimensionKeyError
from ...defaults import DEFAULTS

class FFTDimname():
    '''fft dimension name, along with methods to get post-fft name.
    name: the pre-fft name.
    rad: whether rad==True. Default None.

    self.pre is the pre-fft name.
    self.post is the post-fft name. Alternatively, use str(self) to see the post-fft name.
    '''
    # # # CREATION / INITIALIZATION # # #
    def __init__(self, name, *, rad=None):
        self.name = name
        self.rad = rad

    @classmethod
    def from_post(cls, post, *, rad=None):
        '''create FFTDimname from post-fft name.

        rad: None or bool
            if incompatible with inferred rad, raise DimensionKeyError.
            None --> compatible with any inferred rad; doesn't cause DimensionKeyError.
        post: str
            the post-fft name, used to infer the pre-fft name and possibly also rad.
            if rad is False, post MUST look like 'freq_{dim}'.
            Otherwise:
                if post looks like any DEFAULTS.FFT_FREQ_RAD_DIMNAMES.values(),
                    use name = the associated key (and infer rad=True).
                elif post looks like 'freqrad_{dim}', use name = dim (and infer rad=True).
                elif post looks like 'freq_{dim}', use name = dim (and infer rad=False).
                else, raise DimensionKeyError.        
        '''
        input_rad = rad
        if (rad is not None) and (not rad):  # rad is False
            if post.startswith('freq_'):
                name = post[len('freq_'):]
                inferred_rad = False
            else:
                raise DimensionKeyError(f'rad=False, but post does not start with "freq_". post={post!r}')
        else:  # rad is None or True
            for key, val in DEFAULTS.FFT_FREQ_RAD_DIMNAMES.items():
                if post == val:
                    name = key
                    inferred_rad = True
                    break
            else:  # did not break
                if post.startswith('freqrad_'):
                    name = post[len('freqrad_'):]
                    inferred_rad = True
                elif post.startswith('freq_'):
                    name = post[len('freq_'):]
                    inferred_rad = False
                else:
                    errmsg = (f'cannot determine name for pre-fft dimension associated with post={post!r}.\n'
                            'Expected post="freq_{dim}" or "freqrad_{dim}", or in DEFAULTS.FFT_FREQ_RAD_DIMNAMES.values()')
                    raise DimensionKeyError(errmsg)
        if input_rad is None:
            use_rad = inferred_rad
        else:
            if inferred_rad != input_rad:
                errmsg = (f'input rad={input_rad!r} incompatible with inferred rad={inferred_rad!r}, for post={post!r}')
                raise DimensionKeyError(errmsg)
            use_rad = input_rad
        return cls(name, rad=use_rad)

    @classmethod
    def implied_from(cls, s, array_dims, *, rad=None, post_fft=False):
        '''returns FFTDimname from s corresponding to one of the array_dims.

        s: str
            the pre-fft or post-fft name, corresponding to one of the array_dims.
            Might match an array_dim exactly, or might match array_dim pre or post fft.
        array_dims: iterable of str
            result will be appropriate for taking an fft of array with these dims.
            (Can also contain non-dimension coordinates of array.)
        rad: None or bool
            rad to use when creating FFTDimname from s.
            If None, use rad inferred from s & array_dims.
            Otherwise, ensure rad is compatible with inferred rad (else raise DimensionKeyError).
        post_fft: bool, default False
            True --> array_dims tell the post-fft dimensions.
            False --> array_dims tell the pre-fft dimensions.
            Use True when taking an ifft of an array with array.dims == array_dims.

        Examples:
            ('x', ['x', 'y']) --> cls('x', rad=rad)  # by default, rad=None)
            ('k_x', ['x', 'y']) --> cls('x', rad=True)
            ('freq_x', ['x', 'y']) --> cls('x', rad=False)
            ('x', ['k_x', 'k_y'], post_fft=True) --> cls('x', rad=True)
            ('k_x', ['k_x', 'k_y'], post_fft=True) --> cls('x', rad=True)
            ('freq_x', ['k_x', 'k_y'], post_fft=True) --> raise DimensionKeyError
            ('x', ['k_y', 't']) --> raise DimensionKeyError
        '''
        array_dims = tuple(array_dims)
        input_rad = rad
        attempt_rad = [False, True] if rad is None else [rad]
        if post_fft:
            for d in array_dims:
                for rad in attempt_rad:
                    try:
                        fname = cls.from_post(d, rad=rad)
                    except DimensionKeyError:
                        continue  # can't create fname from d & rad.
                    # else
                    if s == fname.pre:
                        return fname
                    if s == fname.post:
                        return fname
        else:  # post_fft=False. This is the default behavior.
            for d in array_dims:
                fname = cls(d, rad=input_rad)
                if s == fname.pre:  # <-- fname.pre is independent of rad
                    return fname
                for rad in attempt_rad:
                    fname.rad = rad
                    if s == fname.post:
                        return fname
        # <-- didn't find a match.
        errmsg = f'no corresponding dim for s={s!r} in array_dims={array_dims!r}'
        if rad is not None: errmsg += f' when rad={rad!r}'
        if post_fft: errmsg += ' when post_fft=True'
        raise DimensionKeyError(errmsg)
    
    # # # PROPERTIES # # #
    @property
    def pre(self):
        '''the pre-fft name.'''
        return self.name

    @property
    def post(self):
        '''the post-fft name.'''
        if self.rad:
            defaults = DEFAULTS.FFT_FREQ_RAD_DIMNAMES
            if (defaults is None) or (self.name in defaults):
                return defaults[self.name]
            return f'freqrad_{self.name}'
        else:
            return f'freq_{self.name}'

    # # # EQUALITY # # #
    def __eq__(self, other):
        '''self == other. True when self.name == other.name and self.rad == other.rad.
        if other not an instance of type(self), return False.
        '''
        if not isinstance(other, type(self)):
            return False
        return (self.name == other.name) and (self.rad == other.rad)

    # # # MISC. METHODS # # #
    @classmethod
    def is_valid_post(cls, post, *, rad=None):
        '''returns whether post is a valid post-fft name, given rad.
        valid if can create FFTDimname from post & rad without raising DimensionKeyError.
        '''
        try:
            cls.from_post(post, rad=rad)
            return True
        except DimensionKeyError:
            return False

    # # # DISPLAY # # #
    def __str__(self):
        '''str(self). Equivalent to str(self.post).'''
        return str(self.post)

    def __repr__(self):
        '''repr(self).'''
        return f'{type(self).__name__}({self.name!r}, rad={self.rad!r})'



"""
File Purpose: EbysusCollisionsLoader
"""
import xarray as xr

from ...errors import CollisionsModeError
from ...quantities import CollisionsLoader
from ...tools import UNSET, format_docstring

class EbysusCollisionsLoader(CollisionsLoader):
    '''collision frequency calculations.
    allows collisions_mode = 'helita' to get value via helita.
    '''
    @known_var(dims=['snap', 'fluid', 'jfluid'])
    def get_nusj_helita(self):
        '''collision frequency. (directly from Ebysus)
        for a single particle of s (self.fluid) to collide with any of j (self.jfluid).
        '''
        return self.load_maindims_var_across_dims('nu_ij', dims=['snap', 'fluid', 'jfluid'])

    COLLISIONS_MODE_OPTIONS = CollisionsLoader.COLLISIONS_MODE_OPTIONS.copy()
    COLLISIONS_MODE_OPTIONS['helita'] = \
            '''Use nusj_helita to get collision frequency.
            (However, for same fluid & jfluid, use 0 instead.)'''

    COLLISION_TYPE_TO_VAR = CollisionsLoader.COLLISION_TYPE_TO_VAR.copy()
    COLLISION_TYPE_TO_VAR['helita'] = 'nusj_helita'

    @format_docstring(docs_super=CollisionsLoader.get_collision_type)
    def get_collision_type(self):
        '''Similar to super().get_collision_type, but also allows 'helita' <--> do 'helita' collisions.
        
        super().get_collision_type docs copied here for reference:
        ----------------------------------------------------------
        {docs_super}
        '''
        try:
            return super().get_collision_type()
        except CollisionsModeError:
            if self.collisions_mode == 'helita':
                return xr.DataArray('helita')
            else:
                raise
